import unittest

from mygrations.formats.mysql.file_reader.parsers.type_numeric import type_numeric
class test_type_numeric(unittest.TestCase):
    def test_simple(self):

        # parse typical insert values
        parser = type_numeric()
        returned = parser.parse("created int(10) not null default 0,")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('created', parser.name)
        self.assertEquals('INT', parser.column_type)
        self.assertEquals('10', parser.length)
        self.assertFalse(parser.unsigned)
        self.assertFalse(parser.null)
        self.assertFalse(parser.auto_increment)
        self.assertEquals('0', parser.default)
        self.assertTrue(parser.has_comma)
        self.assertEquals(0, len(parser.errors))
        self.assertEquals(0, len(parser.warnings))
        self.assertEquals("`created` INT(10) NOT NULL DEFAULT 0", str(parser))

    def test_auto_increment(self):

        # parse typical insert values
        parser = type_numeric()
        returned = parser.parse("created int(10) not null auto_increment,")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('created', parser.name)
        self.assertEquals('INT', parser.column_type)
        self.assertEquals('10', parser.length)
        self.assertTrue(parser.auto_increment)
        self.assertFalse(parser.unsigned)
        self.assertFalse(parser.null)
        self.assertEquals(None, parser.default)
        self.assertTrue(parser.has_comma)
        self.assertEquals(0, len(parser.errors))
        self.assertEquals(0, len(parser.warnings))
        self.assertEquals("`created` INT(10) NOT NULL AUTO_INCREMENT", str(parser))

    def test_optional_default(self):

        # parse typical insert values
        parser = type_numeric()
        returned = parser.parse("created int(10) UNSIGNED")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('created', parser.name)
        self.assertEquals('INT', parser.column_type)
        self.assertTrue(parser.unsigned)
        self.assertEquals('10', parser.length)
        self.assertTrue(parser.null)
        self.assertEquals(None, parser.default)
        self.assertFalse(parser.has_comma)
        self.assertEquals(0, len(parser.errors))
        self.assertEquals(0, len(parser.warnings))
        self.assertEquals("`created` INT(10) UNSIGNED", str(parser))

    def test_strip_backticks(self):

        # parse typical insert values
        parser = type_numeric()
        returned = parser.parse("`created` int(10) UNSIGNED")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertEquals('created', parser.name)
        self.assertEquals("`created` INT(10) UNSIGNED", str(parser))

    def test_warning_for_string_default(self):

        # parse typical insert values
        parser = type_numeric()
        returned = parser.parse("created int(10) default '0'")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertTrue('does not need to be quoted' in parser.warnings[0])
        self.assertEquals("`created` INT(10) DEFAULT 0", str(parser))

    def test_not_null_needs_default(self):

        # parse typical insert values
        parser = type_numeric()
        returned = parser.parse("name int(10) NOT NULL")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertTrue('is not null and has no default' in parser.warnings[0])

    def test_char_field_defaults_should_be_quoted(self):

        parser = type_numeric()
        returned = parser.parse("name varchar(255) DEFAULT asdf")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertTrue('should be quoted' in parser.warnings[0])

    def test_char_fields_should_not_be_unsigned(self):

        parser = type_numeric()
        returned = parser.parse("name varchar(255) UNSIGNED NOT NULL DEFAULT 'asdf'")

        self.assertTrue(parser.is_char)
        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertTrue('character type and cannot be unsigned' in parser.errors[0])

    def test_char_fields_should_not_auto_increment(self):

        parser = type_numeric()
        returned = parser.parse("name varchar(255) NOT NULL AUTO_INCREMENT")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertTrue('only numeric fields can auto increment' in parser.errors[0])

    def test_autoincrement_can_be_not_null_and_no_default(self):

        parser = type_numeric()
        returned = parser.parse("name int(10) UNSIGNED NOT NULL AUTO_INCREMENT")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertEquals(0, len(parser.warnings))

    def test_number_field_defaults_should_not_be_quoted(self):

        parser = type_numeric()
        returned = parser.parse("name int(10) DEFAULT '0'")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertTrue('does not need to be quoted' in parser.warnings[0])

    # not for everyone
    def test_disallowed_types(self):

        # anything not on the whitelist (this is not a complete list)
        for coltype in ['date', 'bob', 'difjerejrie']:
            parser = type_numeric()
            returned = parser.parse("name %s(10)" % coltype)
            self.assertTrue(
                'not allowed to have a length' in parser.errors[0],
                'Type %s should not be allowed to be numeric but is' % coltype
            )

    # everything on the white list is allowed to be numeric
    def test_valid_column_type(self):

        for coltype in [
            'bit', 'tinyint', 'smallint', 'mediumint', 'int', 'integer', 'bigint', 'decimal', 'numeric', 'char',
            'varchar'
        ]:

            parser = type_numeric()
            returned = parser.parse('name %s(10)' % coltype)

            self.assertTrue(parser.matched)
            self.assertEquals('', returned)
            self.assertEquals(0, len(parser.errors), 'Type %s should not be allowed to have decimals but is' % coltype)



import unittest

from mygrations.formats.mysql.file_reader.database import database as database_reader
from mygrations.formats.mysql.mygrations.mygration import mygration
class test_add_conflicting_fks(unittest.TestCase):

    accounts_table = """CREATE TABLE `accounts` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `name` VARCHAR(255) NOT NULL DEFAULT '',
            PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""

    def _get_db1(self):
        table1 = """CREATE TABLE `tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `account_id` INT(10) UNSIGNED NOT NULL,
            `repeating_task_id` INT(10) UNSIGNED NOT NULL,
            `name` VARCHAR(255) NOT NULL DEFAULT '',
            PRIMARY KEY (`id`),
            KEY `account_id_tasks` (`account_id`),
            KEY `repeating_task_id_tasks` (`repeating_task_id`),
            CONSTRAINT `account_id_tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        return database_reader([table1, self.accounts_table])

    def test_simple_add_and_modify(self):
        """ Migrate to a database that has one extra table and column """

        table1 = """CREATE TABLE `tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `account_id` INT(10) UNSIGNED NOT NULL,
            `repeating_task_id` INT(10) UNSIGNED NOT NULL,
            `name` VARCHAR(255) NOT NULL DEFAULT '',
            `subject` TEXT,
            PRIMARY KEY (`id`),
            KEY `account_id_tasks` (`account_id`),
            KEY `repeating_task_id_tasks` (`repeating_task_id`),
            CONSTRAINT `account_id_tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        table3 = """CREATE TABLE `histories` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
`name` VARCHAR(255) NOT NULL DEFAULT '',
PRIMARY KEY (`id`)) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        db2 = database_reader([table1, self.accounts_table, table3])
        db1 = self._get_db1()

        mygrate = mygration(db2, db1)

        ops = [str(op) for op in mygrate.operations]
        self.assertEquals('SET FOREIGN_KEY_CHECKS=0;', ops[0])
        self.assertEquals(table3, ops[1])
        self.assertEquals('ALTER TABLE `tasks` ADD `subject` TEXT AFTER `name`;', ops[2])
        self.assertEquals('SET FOREIGN_KEY_CHECKS=1;', ops[3])

    def test_add_column_and_mutually_dependent_fk(self):
        """ Add a column to a table that depends upon a table with a mutually-dependent FK """
        table1 = """CREATE TABLE `tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
`account_id` INT(10) UNSIGNED NOT NULL,
`repeating_task_id` INT(10) UNSIGNED NOT NULL,
`name` VARCHAR(255) NOT NULL DEFAULT '',
`subject` TEXT,
PRIMARY KEY (`id`),
KEY `account_id_tasks` (`account_id`),
KEY `repeating_task_id_tasks` (`repeating_task_id`),
CONSTRAINT `account_id_tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
CONSTRAINT `repeating_task_id_tasks_fk` FOREIGN KEY (`repeating_task_id`) REFERENCES `repeating_tasks` (`id`) ON DELETE CASCADE ON UPDATE CASCADE) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        table2 = """CREATE TABLE `repeating_tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
`account_id` INT(10) UNSIGNED NOT NULL,
`task_id` INT(10) UNSIGNED NOT NULL,
`name` VARCHAR(255) NOT NULL DEFAULT '',
PRIMARY KEY (`id`),
KEY `account_id_rts` (`account_id`),
KEY `task_id_rts` (`task_id`),
CONSTRAINT `account_id_tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
CONSTRAINT `task_id_rts` FOREIGN KEY (`task_id`) REFERENCES `tasks` (`id`) ON DELETE CASCADE ON UPDATE CASCADE) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        db2 = database_reader([table1, table2, self.accounts_table])

        mygrate = mygration(db2, self._get_db1())
        ops = [str(op) for op in mygrate.operations]

        self.assertEquals('SET FOREIGN_KEY_CHECKS=0;', ops[0])
        self.assertEquals(table2, ops[1])

        # and then tasks will be modified
        self.assertEquals(
            'ALTER TABLE `tasks` ADD `subject` TEXT AFTER `name`, ADD CONSTRAINT `repeating_task_id_tasks_fk` FOREIGN KEY (`repeating_task_id`) REFERENCES `repeating_tasks` (`id`) ON DELETE CASCADE ON UPDATE CASCADE;',
            ops[2]
        )
        self.assertEquals('SET FOREIGN_KEY_CHECKS=1;', ops[3])

    def test_all_key_adjustments(self):
        """ Add/remove/change keys! """

        table1 = """CREATE TABLE `tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `account_id` INT(10) UNSIGNED NOT NULL,
            `repeating_task_id` INT(10) UNSIGNED NOT NULL,
            `name` VARCHAR(255) NOT NULL DEFAULT '',
            PRIMARY KEY (`id`),
            KEY `account_id_tasks` (`account_id`,`name`),
            KEY `cool_key` (`name`),
            CONSTRAINT `account_id_tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        db2 = database_reader([table1, self.accounts_table])
        db1 = self._get_db1()

        mygrate = mygration(db2, db1)
        ops = [str(op) for op in mygrate.operations]

        self.assertEquals('SET FOREIGN_KEY_CHECKS=0;', ops[0])
        self.assertEquals(
            'ALTER TABLE `tasks` ADD KEY `cool_key` (`name`), DROP KEY `repeating_task_id_tasks`, DROP KEY `account_id_tasks`, ADD KEY `account_id_tasks` (`account_id`,`name`);',
            ops[1]
        )
        self.assertEquals('SET FOREIGN_KEY_CHECKS=1;', ops[2])

    def test_all_column_adjustments(self):
        """ Add/remove/change columns! """

        table1 = """CREATE TABLE `tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `account_id` INT(10) UNSIGNED NOT NULL,
            `name` CHAR(16) DEFAULT NULL,
            `subject` TEXT,
            PRIMARY KEY (`id`),
            KEY `account_id_tasks` (`account_id`),
            CONSTRAINT `account_id_tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        db2 = database_reader([table1, self.accounts_table])
        db1 = self._get_db1()

        mygrate = mygration(db2, db1)
        ops = [str(op) for op in mygrate.operations]

        self.assertEquals('SET FOREIGN_KEY_CHECKS=0;', ops[0])
        self.assertEquals(
            'ALTER TABLE `tasks` ADD `subject` TEXT AFTER `name`, CHANGE `name` `name` CHAR(16), DROP repeating_task_id, DROP KEY `repeating_task_id_tasks`;',
            ops[1]
        )
        self.assertEquals('SET FOREIGN_KEY_CHECKS=1;', ops[2])

    def test_all_constraint_adjustments(self):
        """ Add/remove/change constraints! """

        table1 = """CREATE TABLE `tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `account_id` INT(10) UNSIGNED NOT NULL,
            `task_id` INT(10) UNSIGNED NOT NULL,
            `repeating_task_id` INT(10) UNSIGNED NOT NULL,
            `name` VARCHAR(255) NOT NULL DEFAULT '',
            PRIMARY KEY (`id`),
            KEY `account_id_tasks` (`account_id`),
            KEY `repeating_task_id_tasks` (`repeating_task_id`),
            KEY `task_id` (`task_id`),
            CONSTRAINT `account_id_tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
            CONSTRAINT `task_id_fk` FOREIGN KEY (`task_id`) REFERENCES `tasks` (`id`) ON DELETE CASCADE ON UPDATE CASCADE
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        db1 = database_reader([table1, self.accounts_table])

        table1 = """CREATE TABLE `tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `account_id` INT(10) UNSIGNED NOT NULL,
            `task_id` INT(10) UNSIGNED NOT NULL,
            `repeating_task_id` INT(10) UNSIGNED NOT NULL,
            `name` VARCHAR(255) NOT NULL DEFAULT '',
            PRIMARY KEY (`id`),
            KEY `account_id_tasks` (`account_id`),
            KEY `repeating_task_id_tasks` (`repeating_task_id`),
            KEY `task_id` (`task_id`),
            CONSTRAINT `account_id_tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE RESTRICT ON UPDATE CASCADE,
            CONSTRAINT `task_id_2_fk` FOREIGN KEY (`task_id`) REFERENCES `tasks` (`id`) ON DELETE CASCADE ON UPDATE CASCADE
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        db2 = database_reader([table1, self.accounts_table])

        mygrate = mygration(db2, db1)
        ops = [str(op) for op in mygrate.operations]

        self.assertEquals('SET FOREIGN_KEY_CHECKS=0;', ops[0])
        # foreign key constraints are dropped first in their own operation.  This helps
        # with a couple issues, including the fact that you can't drop a key if a foreign
        # key is using it, and that modifying a foreign key requires dropping it and then
        # re-adding it, but the re-adding *has* to happen in a separate alter command
        self.assertEquals(
            'ALTER TABLE `tasks` DROP FOREIGN KEY `task_id_fk`, DROP FOREIGN KEY `account_id_tasks_fk`;', ops[1]
        )
        self.assertEquals(
            'ALTER TABLE `tasks` ADD CONSTRAINT `task_id_2_fk` FOREIGN KEY (`task_id`) REFERENCES `tasks` (`id`) ON DELETE CASCADE ON UPDATE CASCADE, ADD CONSTRAINT `account_id_tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE RESTRICT ON UPDATE CASCADE;',
            ops[2]
        )
        self.assertEquals('SET FOREIGN_KEY_CHECKS=1;', ops[3])

    def test_no_operations_on_1215(self):

        table1 = """CREATE TABLE `tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `account_id` INT(10) NOT NULL,
            `repeating_task_id` INT(10) UNSIGNED NOT NULL,
            `name` VARCHAR(255) NOT NULL DEFAULT '',
            PRIMARY KEY (`id`),
            KEY `account_id_tasks` (`account_id`),
            KEY `repeating_task_id_tasks` (`repeating_task_id`),
            CONSTRAINT `account_id_tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""

        db = database_reader([table1, self.accounts_table])
        mygrate = mygration(db)

        self.assertEquals([
            'MySQL 1215 error for foreign key `account_id_tasks_fk`: unsigned mistmatch. `accounts`.`id` is unsigned but `tasks`.`account_id` is not'
        ], mygrate.errors_1215)
        self.assertEquals(None, mygrate.operations)



import unittest
from mygrations.formats.mysql.mygrations.operations.row_insert import row_insert
from collections import OrderedDict
class test_row_insert(unittest.TestCase):
    def test_simple(self):
        op = row_insert('a_table', OrderedDict([('id', 5), ('name', 'bob'), ('age', 2)]))

        self.assertEquals("INSERT INTO `a_table` (`id`, `name`, `age`) VALUES ('5', 'bob', '2');", str(op))
        self.assertEquals('a_table', op.table_name)



import unittest

from mygrations.formats.mysql.file_reader.database import database as database_reader
from mygrations.formats.mysql.file_reader.create_parser import create_parser
from mygrations.formats.mysql.mygrations.mygration import mygration
class test_add_conflicting_fks(unittest.TestCase):
    def test_add_conflicting_separates_fks(self):
        """ 3 adds. """

        table1 = """CREATE TABLE `tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
`account_id` INT(10) UNSIGNED NOT NULL,
`repeating_task_id` INT(10) UNSIGNED NOT NULL,
`name` VARCHAR(255) NOT NULL DEFAULT '',
PRIMARY KEY (`id`),
KEY `account_id_tasks` (`account_id`),
KEY `repeating_task_id_tasks` (`repeating_task_id`),
CONSTRAINT `account_id_tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
CONSTRAINT `repeating_task_id_tasks_fk` FOREIGN KEY (`repeating_task_id`) REFERENCES `repeating_tasks` (`id`) ON DELETE CASCADE ON UPDATE CASCADE) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        table2 = """CREATE TABLE `repeating_tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
`account_id` INT(10) UNSIGNED NOT NULL,
`task_id` INT(10) UNSIGNED NOT NULL,
`name` VARCHAR(255) NOT NULL DEFAULT '',
PRIMARY KEY (`id`),
KEY `account_id_rts` (`account_id`),
KEY `task_id_rts` (`task_id`),
CONSTRAINT `account_id_tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
CONSTRAINT `task_id_rts` FOREIGN KEY (`task_id`) REFERENCES `tasks` (`id`) ON DELETE CASCADE ON UPDATE CASCADE) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        table3 = """CREATE TABLE `accounts` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
`name` VARCHAR(255) NOT NULL DEFAULT '',
PRIMARY KEY (`id`)) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        tables = [table1, table2, table3]
        db = database_reader(tables)
        mygrate = mygration(db)

        ops = [str(op) for op in mygrate.operations]

        self.assertEquals('SET FOREIGN_KEY_CHECKS=0;', ops[0])
        self.assertTrue(table1 in ops)
        self.assertTrue(table2 in ops)
        self.assertTrue(table3 in ops)
        self.assertEquals('SET FOREIGN_KEY_CHECKS=1;', ops[4])



import unittest

from mygrations.formats.mysql.db_reader.database import database as database_reader
from mygrations.drivers.mysqldb.mysqldb import mysqldb
from tests.mocks.db.mysql.db_structure import db_structure
from mygrations.formats.mysql.mygrations.row_mygration import row_mygration
class test_mygrate_rows_different_types(unittest.TestCase):
    def _get_tables(self):

        return {
            'logs':
            """
                CREATE TABLE `logs` (
                    `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                    `account_id` int(10) unsigned not null,
                    `message` TEXT NOT NULL,
                    PRIMARY KEY (`id`)
                ) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        }

    def test_modify_rows(self):
        """ Types can change depending on whether or not rows come out of files or the database.  As a result, equality comparison has to ignore type differences """

        # I could also just create the database and add rows but I feel like doing this
        tables = self._get_tables()
        rows = {'logs': ({'id': 1, 'account_id': 1, 'message': 'hey'}, {'id': 2, 'account_id': 1, 'message': 'sup'})}

        database = database_reader(mysqldb(db_structure(tables, rows)))
        database.read_rows('logs')

        rows_from = {
            'logs': ({
                'id': 1,
                'account_id': '1',
                'message': 'hey'
            }, {
                'id': 2,
                'account_id': '1',
                'message': 'sup'
            })
        }
        database_from = database_reader(mysqldb(db_structure(tables, rows_from)))
        database_from.read_rows('logs')

        operations = [str(op) for op in row_mygration(database, database_from).operations]
        self.assertEquals(0, len(operations))



import re
class db_structure:
    """ Mock class used when working with classes which read the database structure/rows

    This handles just a few simple queries: SHOW TABLES, SHOW CREATE TABLE,
    and SELECT * FROM
    """

    def __init__(self, tables, table_rows):

        self.tables = tables
        self.table_rows = table_rows
        self.executed = False

    def cursor(self, cursor_type=None):
        """ Mock for the db `cursor` method.  Returns the equivalent of the default cursor

        cursor_type is ignored but included for consisitency with API standard

        :returns: self
        :rtype: self
        """
        return self

    def execute(self, query):
        """ Checks the query to see what is being executed and loads up the appropriate result set

        raises a ValueError if an unsupported query is found
        :returns: None
        """
        # You could imagine each query being handled by a separate object, but
        # there are only three queries we are supporting, and that is unlikely
        # to change anytime soon, so I'm going to keep this all in the same class
        normalized = re.sub(r'\s+', ' ', query).lower()
        if normalized == 'show tables':
            self._load_show_tables_result()
        elif normalized[:17] == 'show create table':
            self._load_show_create_result(normalized)
        else:
            self._load_select_all_result(normalized)

        self.executed = True
        self.iterator_index = -1

    def _load_show_tables_result(self):
        """ Loads up a result set internally to act as if a SHOW TABLES query was executed
        """
        tables = [val for val in self.tables.keys()]
        tables.sort()

        # results needs to be a list of tuples
        self.results = [(val, ) for val in tables]

    def _load_show_create_result(self, query):
        """ Loads up a result set internally to act as if a SHOW CREATE TABLE query was executed

        :param query: The query being executed
        :type query: string
        """
        table_name = query.split()[-1]
        if not table_name in self.tables:
            raise ValueError(
                "Cannot mock cursor.execute for query %s because table %s was not set" % (query, table_name)
            )

        self.results = [(table_name, self.tables[table_name])]

    def _load_select_all_result(self, query):
        """ Loads up a result set internally to act as if a SELECT * FROM query was executed

        :param query: The query executed
        :type query: string
        """
        m = re.match('select \* from (\S+)', query)
        if not m:
            raise ValueError(
                "Cannot mock cursor.execute for query %s because I was expecting a select query but cannot find the table name"
                % query
            )

        table_name = m.groups()[0].strip('`')
        if not table_name in self.table_rows:
            raise ValueError(
                "Cannot mock cursor.execute for query %s because table %s was not set to have any rows" %
                (query, table_name)
            )

        self.results = self.table_rows[table_name]

    def __iter__(self):
        """ Initializes iteration through the result set
        """
        return self

    def __next__(self):
        """ Returns the next result set from the iterator

        :returns: Tuple from result set
        :rtype: Tuple
        """
        return self.fetchone()

    def fetchone(self):
        """Returns the next result from the result set """
        if not self.executed:
            raise ValueError("Cannot fetch query results before executing")

        self.iterator_index += 1
        if self.iterator_index >= len(self.results):
            raise StopIteration

        return self.results[self.iterator_index]

    def fetchall(self):
        """ Returns the result set """

        return self.results

    def rowcount(self):
        """ Returns the number of results in the result set

        :returns: The number of records in the result set
        :rtype: integer
        """
        return len(self.results)

    def close(self):

        self.executed = False



from mygrations.core.parse.parser import parser
from mygrations.formats.mysql.definitions.column import column
class type_numeric(parser, column):

    allowed_types = {
        'bit': True,
        'tinyint': True,
        'smallint': True,
        'mediumint': True,
        'int': True,
        'integer': True,
        'bigint': True,
        'decimal': True,
        'numeric': True,
        'char': True,
        'varchar': True
    }

    has_comma = False
    is_char = False

    # `created` int(10) unsigned not null default 0 AUTO_INCREMENT
    rules = [{
        'type': 'regexp',
        'value': '[^\(\s\)]+',
        'name': 'name'
    }, {
        'type': 'regexp',
        'value': '\w+',
        'name': 'type'
    }, {
        'type': 'literal',
        'value': '('
    }, {
        'type': 'regexp',
        'value': '\d+',
        'name': 'length'
    }, {
        'type': 'literal',
        'value': ')'
    }, {
        'type': 'literal',
        'value': 'UNSIGNED',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'NOT NULL',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'DEFAULT ([^\(\s\),]+)',
        'name': 'default',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'AUTO_INCREMENT',
        'optional': True
    }, {
        'type': 'literal',
        'value': ',',
        'optional': True,
        'name': 'ending_comma'
    }]

    def process(self):

        self._errors = []
        self._warnings = []
        self.has_comma = True if 'ending_comma' in self._values else False

        self._name = self._values['name'].strip('`')
        self._column_type = self._values['type']
        self._length = self._values['length']
        self._unsigned = True if 'UNSIGNED' in self._values else False
        self._null = False if 'NOT NULL' in self._values else True
        self._default = self._values['default'] if 'default' in self._values else None
        self._auto_increment = True if 'AUTO_INCREMENT' in self._values else False
        self.is_char = self._column_type in ['char', 'varchar']

        # double check unsigned
        if self._unsigned and self.is_char:
            self._errors.append("Column %s is a character type and cannot be unsigned" % self._name)

        # make sense of the default
        if self._default and len(self._default) >= 2 and self._default[0] == "'" and self._default[-1] == "'":
            self._default = self._default.strip("'")
            if not self.is_char:
                self._warnings.append('Default value for numeric column %s does not need to be quoted' % self._name)
        elif self._default:
            if self._default.lower() == 'null':
                self._default = None
            elif self.is_char:
                self._warnings.append('Default value for column %s should be quoted' % self._name)

        if self._default is None and not self._null and not self._auto_increment:
            self._warnings.append(
                'Column %s is not null and has no default: you should set a default to avoid MySQL warnings' %
                (self._name)
            )

        if self._auto_increment and self.is_char:
            self._errors.append(
                'Column %s is an auto increment character field: only numeric fields can auto increment' % (self._name)
            )

        # only a few types of field are allowed to have decimals
        if not self._column_type.lower() in self.allowed_types:
            self._errors.append(
                'Column of type %s is not allowed to have a length for column %s' % (self._column_type, self._name)
            )



import io, os
from .comment_parser import comment_parser
from .create_parser import create_parser
from .insert_parser import insert_parser
class reader(object):
    def __init__(self):

        self._tables = {}
        self._rows = {}
        self._errors = []
        self._warnings = []
        self._matched = False

    @property
    def matched(self):
        """ Public getter.  Returns whether or not the content was successfully parsed

        :returns: Whether or not parsing was successful
        :rtype: bool
        """
        return self._matched

    @property
    def errors(self):
        """ Public getter.  Returns a list of parsing errors

        :returns: A list of parsing errors
        :rtype: list
        """
        return [] if self._errors is None else self._errors

    @property
    def warnings(self):
        """ Public getter.  Returns a list of parsing/table warnings

        :returns: A list of parsing/table warnings
        :rtype: list
        """
        return [] if self._warnings is None else self._warnings

    @property
    def tables(self):
        """ Public getter.  Returns a list of table definitions

        :returns: A list of table definitions
        :rtype: [mygrations.formats.mysql.definitions.table]
        """
        return self._tables

    @property
    def rows(self):
        """ Public getter.  Returns a dictionary containing a lists of rows by table name

        :returns: A dictionary containing list of rows by table name
        :rtype: {table_name: [mygrations.formats.mysql.defintions.row]}
        """
        return self._rows

    """ Helper that returns info about the current filename (if present) for error messages

    :returns: Part of an error message
    :rtype: string
    """

    def _filename_notice(self):

        if self.filename:
            return ' in file %s' % self.filename

        return ''

    """ Reads the file, if necessary

    Reader is a bit more flexible than the other parsers.  It can accept a filename,
    file-like object, or a string.  This method handles that flexibility, taking
    the input from the _parse method and extracting the actual contents no matter
    what was passed in.

    :returns: The data to parse
    :rtype: string
    """

    def _unpack(self, filename):

        # be flexible about what we accept
        # file pointer
        self.filename = ''
        if isinstance(filename, io.IOBase):
            contents = filename.read()

        # and an actual string
        elif isinstance(filename, str):

            # which could be a filename
            if os.path.isfile(filename):
                self.filename = filename
                fp = open(filename, 'r')
                contents = fp.read()
                fp.close()
            else:
                contents = filename

        else:
            raise ValueError(
                "Unknown type for filename: must be an ascii string, a filename, file pointer, or StringIO"
            )

        return contents

    """ Main parsing loop: attempts to find create, insert, and comments in the SQL string
    """

    def parse(self, filename):

        data = self._unpack(filename)

        # okay then!  This is our main parsing loop.
        c = 0
        while data:
            c += 1
            if c > 10000:
                raise ValueError("Exceeded max parse depth")

            # never hurts
            data = data.strip()

            # now we are looking for one of three things:
            # comment, create, insert
            if data[:2] == '--' or data[:2] == '/*' or data[0] == '#':
                parser = comment_parser()
                data = parser.parse(data)

            elif data[:6].lower() == 'create':
                parser = create_parser()
                data = parser.parse(data)
                self._tables[parser.name] = parser

            elif data[:6].lower() == 'insert':
                parser = insert_parser()
                data = parser.parse(data)
                if not parser.table in self._rows:
                    self._rows[parser.table] = []

                self._rows[parser.table].append(parser)

            else:
                self._errors.append("Unrecognized MySQL command: %s%s" % (data, self._filename_notice()))
                return data

            for error in parser.errors:
                self._errors.append('%s%s' % (error, self._filename_notice()))
            for warning in parser.warnings:
                self._warnings.append('%s%s' % (warning, self._filename_notice()))

        self._matched = True
        return data



from collections import OrderedDict
from .rows import rows as rows_definition

from ..mygrations.operations.alter_table import alter_table
from ..mygrations.operations.add_column import add_column
from ..mygrations.operations.change_column import change_column
from ..mygrations.operations.remove_column import remove_column
from ..mygrations.operations.add_key import add_key
from ..mygrations.operations.change_key import change_key
from ..mygrations.operations.remove_key import remove_key
from ..mygrations.operations.add_constraint import add_constraint
from ..mygrations.operations.change_constraint import change_constraint
from ..mygrations.operations.remove_constraint import remove_constraint
from ..mygrations.operations.create_table import create_table
from ..mygrations.operations.row_delete import row_delete
from ..mygrations.operations.row_insert import row_insert
from ..mygrations.operations.row_update import row_update
class table(object):

    _name = ''
    _options = None
    _columns = None
    _indexes = None
    _constraints = None
    _primary = None
    _rows = None
    _errors = None
    _warnings = None
    _auto_increment = 1
    _indexed_columns = None
    _tracking_rows = False

    def __init__(self, name):
        self._name = name
        self._options = []
        self._columns = OrderedDict()
        self._indexes = OrderedDict()
        self._constraints = OrderedDict()
        self._errors = []
        self._warnings = []
        self._primary = ''

    @property
    def name(self):
        """ Public getter.  Returns the name of the table.

        :returns: The table name
        :rtype: string
        """

        return self._name

    @property
    def options(self):
        """ Public getter.  Returns a list of table options

        :returns: Table options
        :rtype: list
        """

        return self._options

    def mark_tracking_rows(self):
        """ Marks the table as having had its rows read, for bookeeping purposes

        The reason this is here is because we keep track of whether or not we are "tracking" rows
        for the current table.  This is for bookeeping purposes, largely as a safe-guard attempt
        to try to identify any more subtle bugs that might creep in.  Normally self._tracking_rows
        gets set to True when we add rows to a table, but if a table is empty then rows will never
        get added, and instead this method must be called to mark the rows as "tracked".
        """
        self._tracking_rows = True
        if self._rows is None:
            self._rows = OrderedDict()

    @property
    def tracking_rows(self):
        """ Public getter.  Returns True/False to denote whether or not this table is tracking row records

        To be clear on the distinction: just about any table might have rows.  However,
        that doesn't mean that the mygration system should be syncing rows for that table.
        self.tracking_rows == True denotes that the mygration system thinks that we
        should probably be syncing rows for this table.

        :returns: Whether or not the mygration system is tracking rows on the table
        :rtype: bool
        """

        return self._tracking_rows

    @property
    def rows(self):
        """ Public getter.  Returns an ordered dictionary of rows by id

        :returns: Table rows
        :rtype: OrderedDict
        """

        return self._rows

    @property
    def columns(self):
        """ Public getter.  Returns an ordered dictionary of table columns

        :returns: Table columns
        :rtype: OrderedDict
        """

        return self._columns

    @property
    def indexes(self):
        """ Public getter.  Returns an ordered dictionary of table indexes

        :returns: Table indexes
        :rtype: OrderedDict
        """

        return self._indexes

    @property
    def constraints(self):
        """ Public getter.  Returns an ordered dictionary of table constraints

        :returns: Table constraints
        :rtype: OrderedDict
        """

        return self._constraints

    @property
    def primary(self):
        """ Public getter.  Returns the index object for the primary key

        :returns: The index object of the primary key column
        :rtype: isinstance(formats.mysql.definitions.index)
        """

        return self._primary

    @property
    def errors(self):
        """ Public getter.  Returns a list of parsing errors

        :returns: A list of parsing errors
        :rtype: list
        """
        return [] if self._errors is None else self._errors

    @property
    def warnings(self):
        """ Public getter.  Returns a list of parsing/table warnings

        :returns: A list of parsing/table warnings
        :rtype: list
        """
        return [] if self._warnings is None else self._warnings

    @property
    def auto_increment(self):
        """ Public getter.  Returns the autoincrement for the table

        :returns: The auto increment
        :rtype: int
        """
        return self._auto_increment

    @property
    def rows(self):
        """ Public getter.  Returns an ordered dictionary with row data by id

        :returns: An ordered dictionary with row data by id
        :rtype: OrderedDict
        """
        return [] if self._rows is None else self._rows

    def add_rows(self, rows):
        """ Adds a mygrations.formats.mysql.definitions.rows object to the table

        The rows object has some flexibility in terms of columns: it doesn't just
        assume that a value is provided for every column in the table.  Rather,
        there can be a list of columns and only those columns have values (which
        supports the equivalent of MySQL INSERT queries which only specify values
        for some columns).

        Rows with errors will not be processed

        :param rows: The rows to process
        :type rows: mygrations.formats.mysql.definitions.rows
        :returns: An error string if an error is encountered, otherwise True/False
        :rtype: bool | string
        """
        if not isinstance(rows, rows_definition):
            raise ValueError(
                'Only objects of class mygrations.formats.mysql.definitions.rows can be added as rows to a table'
            )

        # we can't process guys with errors
        if rows._errors:
            return False

        self._tracking_rows = True
        if self._rows is None:
            self._rows = OrderedDict()

        # the rows object may have a list of columns.  If not use our own list of columns
        # remember that self._columns is an OrderedDict so converting its keys to a list
        # actually preserves the order (which is a requirement for us)
        columns = rows.columns if rows.num_explicit_columns else list(self._columns.keys())

        for values in rows.raw_rows:
            # rows without explicit columns must be checked for matching columns
            if not rows.num_explicit_columns and len(values) != len(columns):
                return 'Insert values has wrong number of values for table %s and row %s' % (self._name, values)

            # we need to know the id of this record, which means we need
            # to know where in the list of values the id column lives
            try:
                id_index = columns.index('id')
                row_id = int(values[id_index])
            except ValueError:
                row_id = self._auto_increment

            self._auto_increment = max(self._auto_increment, row_id + 1)
            if row_id in self._rows:
                return 'Duplicate row id found for table %s and row %s' % (self.name, values)

            if not row_id:
                return 'Invalid row id of %s found for table %s' % (row_id, self.name)

            self._rows[row_id] = OrderedDict(zip(columns, values))

        return True

    def add_raw_row(self, row):
        """ Adds a row into the table as a dictionary instead of a row object

        A bit of repetition here.  This is similar to what happens inside the main
        loop of self.add_rows, but different enough that I'm not trying to combine
        them right this second.

        :param row: The row to add
        :type row: A dictionary
        :returns: An error string if an error is encountered, otherwise True/False
        :rtype: bool | string
        """
        self._tracking_rows = True
        if self._rows is None:
            self._rows = OrderedDict()

        row_id = row['id'] if 'id' in row else self._auto_increment
        self._auto_increment = max(self._auto_increment, row_id + 1)
        if row_id in self._rows:
            return 'Duplicate row id found for table %s and row %s' % (self._name, row)

        # make sure we have a value for every column in the row, and build an OrderedDict
        converted_row = OrderedDict()
        for column in self._columns.keys():
            if column in row:
                converted_row[column] = row[column]
            else:
                converted_row[column] = self._columns[column].default

        self._rows[row_id] = converted_row

    def column_before(self, column_name):
        """ Returns the name of the column that comes before a given row.

        Returns true if the column is at the beginning of the table

        :param column_name: The name of the column to find the position for
        :type column_name: str
        :returns: The name of the column before the given column, or True if at the beginning
        :rtype: string|True
        """
        # this works because we used an OrderedDict
        columns = [col for col in self.columns.keys()]

        if not column_name in columns:
            raise ValueError(
                "Cannot return column before %s because %s does not exist in table %s" %
                (column_name, column_name, self.name)
            )

        index = columns.index(column_name)
        if index == 0:
            return True

        return columns[index - 1]

    def __str__(self):
        return str(self.create())

    def nice(self):
        return str(self.create(True))

    def create(self, nice=False):
        """ Returns a create table operation that can create this table

        :param nice: Whether or not to return a nicely formatted CREATE TABLE command
        :type nice: bool
        :returns: A create table operation
        :rtype: mygrations.operations.create_table
        """
        return create_table(self, nice)

    def to(self, comparison_table, split_operations=False):
        """ Compares two tables to eachother and returns a list of operations which can bring the structure of the second in line with the first

        In other words, this pseudo code will make table have the same structure as comparison_table

        for operation in table.to( comparison_table ):
            table.apply( operation )

        if split_operations is True then a dict of migration operations will be returned to separate
        foreign key operations from everything else.  The dict will have up to three keys: [ 'removed_fks', 'fks', 'kitchen_sink' ]

        'fks' contains the alter statement operation needed to add/change/remove foreign keys
        'kitchen_sink' contains everything else

        If that division of labor seems arbitrary, it isn't: it is separated out that
        way due to the needs of the overall algorithm.

        If split_operations is false then a single alter table operation will be returned that encompasses all changes

        :param comparison_table: A table to find differences with
        :param split_operations: Whether to combine all operations in one alter table or separate them
        :type comparison_table: mygrations.formats.mysql.definitions.table
        :type split_operations: bool
        :returns: A list of operations to apply to table
        :rtype: list[mygrations.formats.mysql.mygrations.operations.*] | dict
        """
        # start with the columns, obviously
        (added_columns, removed_columns, overlap_columns) = self._differences(self.columns, comparison_table.columns)
        (added_keys, removed_keys, overlap_keys) = self._differences(self.indexes, comparison_table.indexes)
        (added_constraints, removed_constraints,
         overlap_constraints) = self._differences(self.constraints, comparison_table.constraints)

        # keeping in mind the overall algorithm, we're going to separate out all changes into three alter statments
        # these are broken up according to the way that the system has to process them to make sure that foreign
        # keys are not violated during the process
        # 1. Adding columns, changing columns, adding keys, changing keys, removing keys, removing foreign keys
        # 2. Adding foreign keys, changing foreign keys
        # 3. Removing columns
        primary_alter = alter_table(self.name)
        for new_column in added_columns:
            primary_alter.add_operation(
                add_column(comparison_table.columns[new_column], comparison_table.column_before(new_column))
            )

        for overlap_column in overlap_columns:
            # it's really easy to tell if a column changed
            if str(self.columns[overlap_column]) == str(comparison_table.columns[overlap_column]):
                continue

            # FALSE POSITIVE CHECK:
            # if one column has collate/character set and the other doesn't, and that is the only difference,
            # then ignore this difference
            if self.columns[overlap_column].is_really_the_same_as(comparison_table.columns[overlap_column]):
                continue

            primary_alter.add_operation(change_column(comparison_table.columns[overlap_column]))

        for removed_column in removed_columns:
            primary_alter.add_operation(remove_column(self.columns[removed_column]))

        # indexes also go in that first alter table
        for new_key in added_keys:
            primary_alter.add_operation(add_key(comparison_table.indexes[new_key]))
        for removed_key in removed_keys:
            primary_alter.add_operation(remove_key(self.indexes[removed_key]))
        for overlap_key in overlap_keys:
            if str(self.indexes[overlap_key]) == str(comparison_table.indexes[overlap_key]):
                continue
            primary_alter.add_operation(change_key(comparison_table.indexes[overlap_key]))

        # removed foreign key constraints get their own alter because that should always happen first
        removed_constraints_alter = alter_table(self.name)
        for removed_constraint in removed_constraints:
            removed_constraints_alter.add_operation(remove_constraint(self.constraints[removed_constraint]))

        # adding/changing/removing foreign keys gets their own alter
        constraints = alter_table(self.name)
        for added_constraint in added_constraints:
            constraints.add_operation(add_constraint(comparison_table.constraints[added_constraint]))
        for overlap_constraint in overlap_constraints:
            if str(self.constraints[overlap_constraint]) == str(comparison_table.constraints[overlap_constraint]):
                continue

            # foreign key constraints are modified by first dropping the constraint and
            # then adding the new one.  However, these two operations cannot happen in the
            # same alter command.  Kinda a pain.  Oh well.
            removed_constraints_alter.add_operation(remove_constraint(self.constraints[overlap_constraint]))
            constraints.add_operation(add_constraint(comparison_table.constraints[overlap_constraint]))

        # now put it all together
        if split_operations:
            operations = {}
            if removed_constraints_alter:
                operations['removed_fks'] = removed_constraints_alter
            if primary_alter:
                operations['kitchen_sink'] = primary_alter
            if constraints:
                operations['fks'] = constraints
        else:
            operations = []
            for operation in constraints:
                primary_alter.add_operation(operation)
            if removed_constraints_alter:
                operations.append(removed_constraints_alter)
            if primary_alter:
                operations.append(primary_alter)

        return operations

    def to_rows(self, from_table=None):
        """ Compares two tables to eachother and returns a list of operations which can bring the rows of this table in line with the other

        It's important to note (and probably important to think through and change eventually)
        that this method has the opposite meaning of `mygrations.formats.mysql.definitions.table.to()`
        That method is called on the `from` table and operations on the (required) `to` table.
        This method is called on the `to` table and can (optionally) be passed in a `from` table.

        :param from_table: A table to find differences with (or None)
        :type from_table: mygrations.formats.mysql.definitions.table
        :returns: A list of operations to apply to table
        :rtype: list[mygrations.formats.mysql.mygrations.operations.*]
        """
        if from_table and not from_table.tracking_rows:
            raise ValueError(
                "Refusing to compare rows for table %s that is not tracking rows.  Technically I can, but this is probably a sign that you are doing something wrong"
                % (self.name)
            )

        (inserted_ids, deleted_ids, updated_ids) = self._differences(from_table.rows if from_table else {}, self.rows)

        operations = []
        for row_id in inserted_ids:
            operations.append(row_insert(self.name, self.rows[row_id]))

        for row_id in deleted_ids:
            operations.append(row_delete(self.name, row_id))

        for row_id in updated_ids:
            # try to be smart and not update if we don't have to
            (inserted_cols, deleted_cols, updated_cols) = self._differences(self.rows[row_id], from_table.rows[row_id])
            differences = False
            for col in updated_cols:
                if not self._loose_equal(self.rows[row_id][col], from_table.rows[row_id][col]):
                    differences = True
                    break

            if differences:
                operations.append(row_update(self.name, self.rows[row_id]))

        return operations

    def column_is_indexed(self, column):
        """ Returns True/False to denote whether or not the column has a useable index

        :param column: The column to check
        :type column: string|mygrations.formats.mysql.definitions.column
        """
        if type(column) != str:
            column = column.name

        if column not in self._columns:
            return False

        # best way to do this is with a set.  We'll keep a record of all indexed columns
        # the column is indexed if an index has that column in the first position
        if self._indexed_columns is None:
            self._indexed_columns = set([index.columns[0] for index in self._indexes.values()])

        return column in self._indexed_columns

    def _differences(self, from_list, to_list):
        """
        Calculates the difference between two OrderedDicts.

        https://codereview.stackexchange.com/a/176303/140581

        :param from_list: OrderedDict
        :param to_list: OrderedDict
        :return: (added, removed, overlap)
        """

        return ([key for key in to_list if key not in from_list], [key for key in from_list if key not in to_list],
                [key for key in from_list if key in to_list])

    def apply_operation(self, operation):
        """
        Applies an operation to the table

        :param operation: The operation to apply
        :type operation: mygrations.formats.mysql.mygration.operations.*
        """
        operation.apply_to_table(self)

    def add_column(self, column, position=False):
        """ Adds a column to the table

        The value of position matches self.position from mygrations.formats.mysql.mygration.operations.add_column

        :param column: The column to add
        :param position: Where to put the column
        :type column: mygrations.formats.mysql.definitions.column
        :type position: See mygrations.formats.mysql.mygration.operations.add_column
        """
        # putting it at the end is easy
        if column.name in self._columns:
            raise ValueError("Cannot add column %s because %s already exists" % (column.name, column.name))

        if not position:
            self._columns[column.name] = column
            return True

        # end is also easy
        if position == True:
            self._columns[column.name] = column
            self._columns.move_to_end(column.name, last=False)
            return True

        # now it is tricky
        found = False
        new_columns = OrderedDict()
        for key, value in self._columns.items():
            new_columns[key] = value
            if key == position:
                new_columns[column.name] = column
                found = True

        if not found:
            raise ValueError(
                "Cannot add column %s after %s because %s does not exist" % (column.name, position, position)
            )

        self._columns = new_columns
        return True

    def remove_column(self, column_name):
        """ Removes a column from the table

        :param column_name: The column to remove
        :type column_name: string|mygrations.formats.mysql.definitions.column
        """
        if type(column_name) != str:
            column_name = column_name.name

        if not column_name in self._columns:
            raise ValueError("Cannot remove column %s because column %s does not exist" % (column_name, column_name))

        self._columns.pop(column_name, None)

    def change_column(self, new_column):
        """ Changes a column

        This does not currently support renaming

        :param new_column: The new column definition
        :type new_column: mygrations.formats.mysql.definitions.column
        """
        if not new_column.name in self._columns:
            raise ValueError(
                "Cannot modify column %s because column %s does not exist" % (new_column.name, new_column.name)
            )
        self._columns[new_column.name] = new_column

    def add_key(self, key):
        """ Adds an key to the table

        :param key: The key to add
        :type key: mygrations.formats.mysql.definitions.key
        """
        if key.name in self._indexes:
            raise ValueError("Cannot add key %s because key %s already exists" % (key.name, key.name))
        self._indexes[key.name] = key

        if self._indexed_columns is not None:
            self._indexed_columns.add(key.columns[0])

    def remove_key(self, key):
        """ Removes an key from the table

        :param key: The key to remove
        :type key: string|mygrations.formats.mysql.definitions.key
        """
        if type(key) != str:
            key = key.name

        if key not in self._indexes:
            raise ValueError("Cannot remove key %s because key %s does not exist" % (key, key))

        indexed_column = self._indexes[key].columns[0]
        self._indexes.pop(key, None)

        if self._indexed_columns is not None:
            self._indexed_columns.discard(indexed_column)

    def change_key(self, new_key):
        """ Changes a key

        This does not currently support renaming

        :param new_key: The new key definition
        :type new_key: mygrations.formats.mysql.definitions.key
        """
        if not new_key.name in self._indexes:
            raise ValueError("Cannot modify key %s because key %s does not exist" % (new_key.name, new_key.name))

        if self._indexed_columns is not None:
            self._indexed_columns.discard(self._indexes[new_key.name].columns[0])

        self._indexes[new_key.name] = new_key

        if self._indexed_columns is not None:
            self._indexed_columns.add(new_key.columns[0])

    def add_constraint(self, constraint):
        """ Adds a constraint to the table

        :param constraint: The constraint to add
        :type constraint: mygrations.formats.mysql.definitions.constraint
        """
        if constraint.name in self._constraints:
            raise ValueError(
                "Cannot add constraint %s because constraint %s already exists" % (constraint.name, constraint.name)
            )
        self._constraints[constraint.name] = constraint

    def remove_constraint(self, constraint):
        """ Removes an constraint from the table

        :param constraint: The constraint to remove
        :type constraint: string|mygrations.formats.mysql.definitions.constraint
        """
        if type(constraint) != str:
            constraint = constraint.name

        if constraint not in self._constraints:
            raise ValueError(
                "Cannot remove constraint %s because constraint %s does not exist" % (constraint, constraint)
            )
        self._constraints.pop(constraint, None)

    def change_constraint(self, new_constraint):
        """ Changes a constraint

        This does not currently support renaming

        :param new_constraint: The new constraint definition
        :type new_constraint: mygrations.formats.mysql.definitions.constraint
        """
        if not new_constraint.name in self._constraints:
            raise ValueError(
                "Cannot modify constraint %s because constraint %s does not exist" %
                (new_constraint.name, new_constraint.name)
            )
        self._constraints[new_constraint.name] = new_constraint

    def _loose_equal(self, val1, val2):
        """ Performs a looser comparison, as values might have different types depending on whether they came out of a database or file

        Returns true if the two values are equal, even if one is a string and the other an int.

        :param val1: The first value to compare
        :param val2: The second value to compare
        :type val1: string|int
        :type val2: string|int
        :return: Whether or not they are considered a match
        :rtype: bool
        """
        # if we don't have a type mistmatch then this is easy
        if type(val1) == type(val2):
            return val1 == val2

        # otherwise see if we can cheat and just convert to strings
        return str(val1) == str(val2)



class index(object):

    _name = ''
    _index_type = ''
    _columns = None
    _errors = None
    _warnings = None

    def __init__(self, name, columns, index_type='INDEX'):
        """ Index constructor

        :param name: The name of the index
        :param columns: The columns in the index
        :param index_type: The type of the index (INDEX, UNIQUE, or PRIMARY)
        :type name: string
        :type columns: [string]
        :type index_type: string
        """
        self._name = name
        self._index_type = index_type
        self._columns = columns

    @property
    def name(self):
        """ Public getter.  Returns the name of the column.

        :returns: The index name
        :rtype: string
        """

        return self._name

    @property
    def index_type(self):
        """ Public getter.  Returns a string denoting the type of the index.  Always returns in uppercase

        Index type can be 'INDEX', 'UNIQUE', or 'PRIMARY'

        :returns: The index type
        :rtype: string
        """
        return self._index_type.upper()

    @property
    def columns(self):
        """ Public getter.  Returns a list of the column names on the index.

        :returns: The column length
        :rtype: list
        """

        return self._columns

    @property
    def errors(self):
        """ Public getter.  Returns a list of parsing errors

        :returns: A list of parsing errors
        :rtype: list
        """
        return [] if self._errors is None else self._errors

    @property
    def warnings(self):
        """ Public getter.  Returns a list of parsing/table warnings

        :returns: A list of parsing/table warnings
        :rtype: list
        """
        return [] if self._warnings is None else self._warnings

    def __str__(self):
        """ Returns the MySQL command that would create the column

        i.e. column_name type(len) default ''

        :returns: A partial MySQL command that could be used to generate the column
        :rtype: string
        """
        parts = []
        if self.index_type == 'PRIMARY':
            parts.append('PRIMARY')
        elif self.index_type == 'UNIQUE':
            parts.append('UNIQUE')
        parts.append('KEY')

        if self.name:
            parts.append('`%s`' % self.name)
        parts.append('(`%s`)' % ("`,`".join(self.columns)))

        return ' '.join(parts)



from .constraint_foreign import constraint_foreign
from .index_key import index_key
from .index_primary import index_primary
from .index_unique import index_unique
from .type_character import type_character
from .type_decimal import type_decimal
from .type_enum import type_enum
from .type_numeric import type_numeric
from .type_plain import type_plain
from .type_text import type_text
from .table_option import table_option
from .insert_values import insert_values

__all__ = [
    "constraint_foreign", "index_key", "index_primary", "index_unique", "type_character", "type_decimal", "type_enum",
    "type_numeric", "type_plain", "type_text", "table_option", "insert_values"
]



import os
import glob

from .reader import reader as sql_reader
from mygrations.formats.mysql.definitions.database import database as database_definition
class database(database_definition):
    def __init__(self, strings):
        """ Constructor.  Accepts a string or list of strings with different possible contents

        Strings can be one of the following:

        ==================  ====================
        Type                Value
        ==================  ====================
        string              SQL to parse
        string              A filename to read and to parse as SQL
        string              A directory name to search for .sql files, parsing each one
        list                A list of strings, with each element corresponding to any of the above
        ==================  ====================

        :param strings: A string or list of strings corresponding to one of the allowed input types
        :type strings: string|list
        """

        self._warnings = []
        self._errors = []
        self._tables = {}
        self._rows = []

        if isinstance(strings, str):
            strings = [strings]

        for string in strings:
            self.process(string)

        self.store_rows_with_tables()

    def process(self, string):
        """ Processes a string.

        Strings can be either SQL to parse, the location of an SQL file, or a directory containing SQL files

        :param string: A string containing one of the above
        :type string: string
        """

        if os.path.isdir(string):
            self._process_directory(string)
        elif os.path.isfile(string):
            self._read(string)
        else:
            self._read(string)

    def _process_directory(self, directory):
        """ Processes a directory.

        Finds all SQL files in the directory and calls `_read()` on them,
        which results in the file being parsed and its tables/rows added to the
        record of database tables/rows.

        :param string: A string containing one of the above
        :type string: string
        """

        if directory[-1] != os.sep:
            directory += os.sep

        for filename in glob.glob('%s*.sql' % directory):
            self._read(filename)

    def _read(self, contents):
        """ Processes a file or string of SQL.

        Creates a reader object (which accepts files or a string of SQL)
        to parse its input and stores the tables/rows in the database
        object.

        :param contents: A string containing a filename or SQL
        :type contents: string
        """

        try:
            reader = sql_reader()
            reader.parse(contents)

        except ValueError as e:
            print("Error in file %s: %s" % (contents, e))

        # pull in all errors and warnings
        self._errors.extend(reader.errors)
        self._warnings.extend(reader.warnings)

        # keep rows and tables separate while we are reading
        for (table_name, table) in reader.tables.items():
            if table.name in self._tables:
                self.errors.append('Found two definitions for table %s' % table.name)
            self._tables[table.name] = table

        for (table_name, rows) in reader.rows.items():
            self._rows.extend(rows)



import unittest

from mygrations.formats.mysql.db_reader.database import database as database_reader
from mygrations.formats.mysql.file_reader.database import database as file_reader
from mygrations.drivers.mysqldb.mysqldb import mysqldb
from tests.mocks.db.mysql.db_structure import db_structure
from mygrations.formats.mysql.mygrations.row_mygration import row_mygration
class test_backslashes(unittest.TestCase):
    def test_diffs_with_quotes(self):
        """ Things that need backslashes can cause trouble """

        # stick close to our use case: get the comparison table from the "database"
        tables = {
            'logs':
            """
                CREATE TABLE `logs` (
                    `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                    `message` TEXT NOT NULL,
                    `traceback` text,
                    PRIMARY KEY (`id`)
                ) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        }

        rows = {
            'logs': ({
                'id': 1,
                'message': 'test\\sup',
                'traceback': 'never'
            }, {
                'id': 2,
                'message': 'sup\\test',
                'traceback': 'forever'
            })
        }

        db_db = database_reader(mysqldb(db_structure(tables, rows)))
        db_db.read_rows('logs')

        # and one from a file
        table1 = """CREATE TABLE `logs` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `message` TEXT NOT NULL,
            `traceback` text,
            PRIMARY KEY (`id`)
        ) ENGINE=InnoDB DEFAULT CHARSET=utf8;

        INSERT INTO logs (id,message,traceback) VALUES (1,'test\\sup', 'never');
        INSERT INTO logs (id,message,traceback) VALUES (2,'bob\\test', 'forever');
        """
        files_db = file_reader([table1])

        mygrate = row_mygration(files_db, db_db)
        ops = [str(op) for op in mygrate]
        self.assertEquals(1, len(ops))
        self.assertTrue("`message`='bob\\\\test'" in ops[0])



import unittest

from mygrations.helpers.dotenv import dotenv
import io
import tempfile
class test_dotenv_get_contents(unittest.TestCase):

    dotenv = None
    test_string = 'test string'

    def setUp(self):

        self.dotenv = dotenv()

    # get_contents should accept a number of parameters.
    # It should accept a stringIO wrapper
    def test_get_contents_stringIO(self):

        self.assertEquals(self.test_string, self.dotenv.get_contents(io.StringIO(self.test_string)))

    # it should also accept an actual string
    def test_get_contents_string(self):

        self.assertEquals(self.test_string, self.dotenv.get_contents(self.test_string))

    # as well as a more general file pointer
    def test_get_contents_fp(self):

        fp = tempfile.TemporaryFile()
        fp.write(self.test_string.encode(encoding='UTF-8'))
        fp.seek(0)
        self.assertEquals(self.test_string, self.dotenv.get_contents(fp))
        fp.close()

    # it should also accept a filename
    def test_get_contents_filename(self):

        filename = '%s/unit_mygrations_dotenv' % tempfile.gettempdir()
        fp = open(filename, 'w')
        fp.write(self.test_string)
        fp.close()

        self.assertEquals(self.test_string, self.dotenv.get_contents(filename))



from .rule_base import rule_base
class rule_delimited(rule_base):
    """ rule = rule_delimited( parser, rule, next_rule )

    rule_delimited constructor.  Pass in the parser object that the rule is for, the rule
    that is being parsed, and the next rule in the parsing chain.

    This class takes as input the character that values are delimited by and an optional
    character to use to quote strings.  It will then parse the input string looking for
    a list of values separated by the given delimiter.

    As part of the logic, all delimited rules must be followed up bby a rule_literal object.

    Allowed keys in the rules dictionary:

    =========  ========
    key        contents
    =========  ========
    name       The name of the rule
    separator  The character that separates values
    quote      (optional) A character that encloses values
    optional   (optional) If true, denotes that this rule is optional
    =========  ========

    :param parser: The parser that this rule belongs to
    :param rule: A dictionary representing the rule configuration (see each rule class for details)
    :param next_rule: A dictionary representing the next rule
    :type parser: parser
    :type rule: dict
    :type next_rule: dict
    :returns: rule object
    """

    require_value = False

    separator = ''
    quote = ''

    def __init__(self, parser, rule, next_rule):

        super().__init__(parser, rule, next_rule)

        if not 'separator' in rule or not rule['separator']:
            raise ValueError('Missing separator for rule %s' % rule)

        # assumption: delimiteds are always followed up by a literal
        if not self.next_rule or self.next_rule['type'] != 'literal':
            raise ValueError(
                "Delimited rules must be followed by a literal for rule %s in class %s, next: %s" %
                (self.rule, self.parser_class, self.next_rule)
            )

        if not 'value' in self.next_rule or not self.next_rule['value']:
            raise ValueError(
                "Delimited rules must be followed by a literal rule with a value for rule %s in class %s, next: %s" %
                (self.rule, self.parser_class, self.next_rule)
            )

        self.separator = self.rule['separator']
        self.quote = self.rule['quote'] if 'quote' in self.rule else ''

        # separator and quote must be a single character
        if len(self.separator) > 1:
            raise ValueError(
                "separator for delimited rule must be one character only for rule %s in class %s" %
                (self.rule, self.parser_class)
            )

        if len(self.quote) > 1:
            raise ValueError(
                "quote for delimited rule must be one character only for rule %s in class %s" %
                (self.rule, self.parser_class)
            )

    def parse(self, string):

        end = self.next_rule['value'].lower()
        end_len = len(end)

        # it's ugly, but we will do this one character at a time
        has_word = False
        vals = []
        current = ''
        in_quote = False
        for (index, char) in enumerate(string):

            # if we hit a quote then we need to start/finish a word
            if self.quote and char == self.quote:

                # open word
                if not in_quote:
                    in_quote = True
                    has_word = True
                    current = ''

                # close word
                else:
                    vals.append(current)
                    has_word = False
                    in_quote = False

                continue

            # next deal with other characters outside of quotes
            if not in_quote:

                # are we done parsing?

                if string[index:index + end_len].lower() == end:
                    if has_word:
                        vals.append(current)
                    break

                # if we hit a separator then we need to start a new word
                if char == self.separator:

                    if has_word:
                        vals.append(current)
                        has_word = False
                    current = ''

                    continue

                # skip white space outside of quotes (also end a word)
                if char == ' ':
                    if has_word:
                        vals.append(current)
                        has_word = False

                    current = ''
                    continue

                # otherwise we have a character for a word
                has_word = True
                current += char
                continue

            # If we get here then we are inside a quote and the character
            # we are working with is not a quote.  This is easy.
            current += char

        # We are out of our processing loop!
        # if we are still in quote then there was a syntax error in the SQL
        if in_quote:
            raise SyntaxError("Opening quote without closing quote at %s" % string)

        # did we match anything?
        if not vals:
            self.leftovers = string
            self.result = ''
            return False

        # almost done: one last step.  Remove everything that we parsed from string.
        # we know what index we stopped at, so it is easy.
        self.leftovers = string[index:]
        self.result = vals
        return True



class create_table:
    """ Generates an SQL command to create a table """

    def __init__(self, table, nice=False):
        """ Create table constructor

        :param table: The table to build a create table command for
        :param nice: Whether or not to return a nicely formatted CREATE TABLE command
        :type table: formats.mysql.definitions.table
        :type nice: bool
        """
        self.table = table
        self._nice = nice

    @property
    def table_name(self):
        """ Public getter.  Returns the name of the table.

        :returns: The table name
        :rtype: string
        """

        return self.table.name

    def __str__(self):
        newline = '\n' if self._nice else ''
        padding = '  ' if self._nice else ''
        body = ['%s%s' % (padding, str(self.table.columns[col])) for col in self.table.columns]
        body.extend(['%s%s' % (padding, str(self.table.indexes[index])) for index in self.table.indexes])
        body.extend([
            '%s%s' % (padding, str(self.table.constraints[constraint])) for constraint in self.table.constraints
        ])
        options = ['%s=%s' % (opt.name, opt.value) for opt in self.table.options]
        if options:
            options = ' %s' % ' '.join(options)
        else:
            options = ''
        return 'CREATE TABLE `%s` (%s%s%s)%s;' % (self.table.name, newline, ",\n".join(body), newline, options)



from mygrations.core.parse.parser import parser
from mygrations.formats.mysql.definitions.column import column
class type_decimal(parser, column):

    allowed_types = {'real': True, 'double': True, 'float': True, 'decimal': True, 'numeric': True}

    decimals = ''
    has_comma = False

    # longitude float(20,4) unsigned default null
    rules = [{
        'type': 'regexp',
        'value': '[^\(\s\)]+',
        'name': 'name'
    }, {
        'type': 'regexp',
        'value': '\w+',
        'name': 'type'
    }, {
        'type': 'literal',
        'value': '('
    }, {
        'type': 'regexp',
        'value': '\d+',
        'name': 'length'
    }, {
        'type': 'literal',
        'value': ',',
        'name': 'separating_comma'
    }, {
        'type': 'regexp',
        'value': '\d+',
        'name': 'decimals'
    }, {
        'type': 'literal',
        'value': ')'
    }, {
        'type': 'literal',
        'value': 'UNSIGNED',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'NOT NULL',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'DEFAULT ([^\(\s\),]+)',
        'optional': True,
        'name': 'default'
    }, {
        'type': 'literal',
        'value': ',',
        'optional': True,
        'name': 'ending_comma'
    }]

    def process(self):

        self.has_comma = True if 'ending_comma' in self._values else False

        self._errors = []
        self._warnings = []
        self._name = self._values['name'].strip('`')
        self._column_type = self._values['type']
        self.decimals = self._values['decimals']
        self._length = '%s,%s' % (self._values['length'], self._values['decimals'])
        self._unsigned = True if 'UNSIGNED' in self._values else False
        self._null = False if 'NOT NULL' in self._values else True
        self._default = self._values['default'] if 'default' in self._values else None

        # make sense of the default
        if self._default and len(self._default) >= 2 and self._default[0] == "'" and self._default[-1] == "'":
            self._default = self._default.strip("'")
            self._warnings.append('Default value for numeric column %s does not need to be quoted' % self._name)
        elif self._default and self._default.lower() == 'null':
            self._default = None

        if self._default is None and not self._null:
            self._warnings.append(
                'Column %s is not null and has no default: you should set a default to avoid MySQL warnings' %
                (self._name)
            )

        # only a few types of field are allowed to have decimals
        if not self._column_type.lower() in self.allowed_types:
            self._errors.append(
                'Column of type %s is not allowed to have a decimal length for column %s' %
                (self._column_type, self._name)
            )



import unittest

from mygrations.formats.mysql.file_reader.parsers.type_decimal import type_decimal
class test_type_decimal(unittest.TestCase):
    def test_simple(self):

        # parse typical insert values
        parser = type_decimal()
        returned = parser.parse("latitude float(10,7) not null default 0,")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('column', parser.definition_type)
        self.assertEquals('latitude', parser.name)
        self.assertEquals('FLOAT', parser.column_type)
        self.assertEquals('10,7', parser.length)
        self.assertFalse(parser.unsigned)
        self.assertFalse(parser.null)
        self.assertEquals('0', parser.default)
        self.assertTrue(parser.has_comma)
        self.assertEquals(0, len(parser.errors))
        self.assertEquals(0, len(parser.warnings))
        self.assertTrue(parser.character_set is None)
        self.assertTrue(parser.collate is None)
        self.assertEquals("`latitude` FLOAT(10,7) NOT NULL DEFAULT 0", str(parser))

    def test_optional_default(self):

        # parse typical insert values
        parser = type_decimal()
        returned = parser.parse("latitude float(10,7) UNSIGNED")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('column', parser.definition_type)
        self.assertEquals('latitude', parser.name)
        self.assertEquals('FLOAT', parser.column_type)
        self.assertTrue(parser.unsigned)
        self.assertEquals('10,7', parser.length)
        self.assertTrue(parser.null)
        self.assertEquals(None, parser.default)
        self.assertFalse(parser.has_comma)
        self.assertEquals(0, len(parser.errors))
        self.assertEquals(0, len(parser.warnings))
        self.assertEquals("`latitude` FLOAT(10,7) UNSIGNED", str(parser))

    def test_strip_backticks(self):

        # parse typical insert values
        parser = type_decimal()
        returned = parser.parse("`latitude` float(10,7)")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('column', parser.definition_type)
        self.assertEquals('latitude', parser.name)
        self.assertEquals("`latitude` FLOAT(10,7)", str(parser))

    def test_warning_for_string_default(self):

        # parse typical insert values
        parser = type_decimal()
        returned = parser.parse("latitude float(10,7) default '0'")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertTrue('does not need to be quoted' in parser.warnings[0])

    def test_not_null_needs_default(self):

        # parse typical insert values
        parser = type_decimal()
        returned = parser.parse("name float(10,7) NOT NULL")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertTrue('is not null and has no default' in parser.warnings[0])

    # only some can have decimals
    def test_disallowed_collate(self):

        # parse typical insert values
        for coltype in ['real', 'double', 'float', 'decimal', 'numeric']:
            parser = type_decimal()
            returned = parser.parse("name %s(10,7)" % coltype)
            self.assertEquals(0, len(parser.errors), 'Type %s should be allowed to have decimals but is not' % coltype)

    # everything else is not allowed to have a length
    def test_invalid_column_type(self):

        for coltype in [
            'date', 'year', 'tinyblob', 'blob', 'mediumblob', 'longblob', 'tinytext', 'text', 'mediumtext', 'longtext',
            'json', 'char', 'varchar', 'biejfiejrei'
        ]:

            parser = type_decimal()
            returned = parser.parse('name %s(10,7)' % coltype)

            self.assertTrue(parser.matched)
            self.assertEquals('', returned)
            self.assertTrue(
                'not allowed to have a decimal length' in parser.errors[0],
                'Type %s should not be allowed to have decimals but is' % coltype
            )



from .plan import plan
from mygrations.drivers.mysqldb.mysqldb import mysqldb
def execute(options):

    obj = apply(options)
    obj.execute()
class apply(plan):
    def execute(self):

        commands = self.build_commands()
        connection = mysqldb(self.credentials)
        connection.execute(['%s' % command for command in commands])



import unittest

from mygrations.formats.mysql.file_reader.parsers.insert_values import insert_values
class test_insert_values(unittest.TestCase):
    def test_simple(self):

        # parse typical insert values
        parser = insert_values()
        returned = parser.parse("(      'name', 'bob', 'okay', 1),")

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have some data now
        self.assertEquals('', returned)

        # all we really have is the list of values
        self.assertEquals(['name', 'bob', 'okay', '1'], parser.values)
        self.assertTrue(parser.has_comma)

    def test_optional_comma(self):

        # parse typical insert values
        parser = insert_values()
        returned = parser.parse("('name', 'bob', 'okay', 1)")

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have some data now
        self.assertEquals('', returned)

        # all we really have is the list of values
        self.assertEquals(['name', 'bob', 'okay', '1'], parser.values)
        self.assertFalse(parser.has_comma)

    def test_return(self):

        # parse typical insert values
        parser = insert_values()
        returned = parser.parse("('name','bob','okay',1), ('name','bob')")

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have some data now
        self.assertEquals("('name','bob')", returned)

        # all we really have is the list of values
        self.assertEquals(['name', 'bob', 'okay', '1'], parser.values)
        self.assertTrue(parser.has_comma)



import os
import glob

from ..file_reader.reader import reader as sql_reader
from mygrations.formats.mysql.definitions.database import database as database_definition
class database(database_definition):
    def __init__(self, conn):
        """ Constructor.  Accepts a mygrations db wrapper

        :param conn: mygrations db wrapper
        :type conn: mygrations.drivers.mysqldb.mysqldb
        """
        self.conn = conn
        self._warnings = []
        self._errors = []
        self._tables = {}
        self._rows = []

        # _load_tables will get the party started
        self._load_tables(self.conn)

    def _load_tables(self, conn):
        """ Reads a database from the MySQL connection.

        Accepts a mygrations db wrapper

        :param conn: mygrations db wrapper
        :type conn: mygrations.drivers.mysqldb.mysqldb
        """
        for (table, create_table) in conn.tables().items():
            try:
                reader = sql_reader()
                reader.parse(create_table)

            except ValueError as e:
                print("Error in table %s: %s" % (create_table, e))

            # we shouldn't get any errors, of course, because
            # this is coming out of MySQL.  However, it may still
            # get some warnings (due to lint settings (maybe)).  Either way,
            # keep around the errors just because, even if it is
            # always empty.
            self._errors.extend(reader.errors)
            self._warnings.extend(reader.warnings)

            for (table_name, table) in reader.tables.items():
                if table.name in self._tables:
                    self.errors.append('Found two definitions for table %s' % table.name)

                # our reader will return table objects
                # from the file_reader namespace.  These expect
                # rows to come up inside the SQL that we pass in.
                # However, we don't have any inserts in our SQL.  Nor
                # do I want to just load up all rows and pass them
                # in for storage, because only a small minority of tables
                # will actually be tracking database rows.  Instead
                # I will sort out table records later
                self._tables[table.name] = table

    def read_rows(self, table):
        """ Extracts the rows for the table from the database and stores them in the table object

        This object connects to the actual database.  However, we only need to know
        about the rows on a table in a minority of cases.  Mainly, whenever a table's rows
        are being managed by the mygration system.  This primarily just happens for tables
        which contain configuration information.  How do we know that a table's rows
        are being managed?  **We** don't, but the mygration system does.  If the *.sql file
        for a table has rows then the system assumes that its rows are being managed,
        and when appropriate that will prompt a call to this method which will actually
        load the rows to see if anything has changed.

        :param table: The table to read rows for
        :type table: string|mygrations.formats.mysql.definitions.table
        """
        if type(table) != str:
            table = table.name

        if not table in self.tables:
            raise ValueError("Cannot read rows for table %s because that table is not found in the database object")

        for row in self.conn.rows(table):
            self.tables[table].add_raw_row(row)

        # if the table is empty the system won't realize that we loaded rows for it
        # for bookeeping purposes, mark the table as read
        self.tables[table].mark_tracking_rows()



from mygrations.core.parse.parser import parser
from mygrations.formats.mysql.definitions.column import column
class type_enum(parser, column):

    allowed_types = {'set': True, 'enum': True}

    values = []
    has_comma = False

    # types enum( `young`,`middle`,`old` )
    rules = [{
        'type': 'regexp',
        'value': '[^\(\s\)]+',
        'name': 'name'
    }, {
        'type': 'regexp',
        'value': '\w+',
        'name': 'type'
    }, {
        'type': 'literal',
        'value': '('
    }, {
        'type': 'delimited',
        'name': 'values',
        'quote': "'",
        'separator': ','
    }, {
        'type': 'literal',
        'value': ')'
    }, {
        'type': 'literal',
        'value': 'NOT NULL',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'DEFAULT ([^\(\s\),]+)',
        'optional': True,
        'name': 'default'
    }, {
        'type': 'regexp',
        'value': 'CHARACTER SET ([^\(\s\),]+)',
        'name': 'character_set',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'COLLATE ([^\(\s\),]+)',
        'name': 'collate',
        'optional': True
    }, {
        'type': 'literal',
        'value': ',',
        'optional': True,
        'name': 'ending_comma'
    }]

    def __init__(self, rules=[]):

        super().__init__(rules)

        self._errors = []
        self._warnings = []
        self.values = []

    def process(self):

        self.has_comma = True if 'ending_comma' in self._values else False

        self._name = self._values['name'].strip('`')
        self._column_type = self._values['type']
        self.values = self._values['values']
        self._length = self.values
        self._null = False if 'NOT NULL' in self._values else True
        self._default = self._values['default'] if 'default' in self._values else None
        self._character_set = self._values['character_set'] if 'character_set' in self._values else None
        self._collate = self._values['collate'] if 'collate' in self._values else None

        if self._character_set and len(self._character_set
                                       ) >= 2 and self._character_set[0] == "'" and self._character_set[-1] == "'":
            self._character_set = self._character_set.strip("'")

        if self._collate and len(self._collate) >= 2 and self._collate[0] == "'" and self._collate[-1] == "'":
            self._collate = self._collate.strip("'")

        # make sense of the default
        if self._default and len(self._default) >= 2 and self._default[0] == "'" and self._default[-1] == "'":
            self._default = self._default.strip("'")
        elif self._default:
            if self._default.lower() == 'null':
                self._default = None
            else:
                self._warnings.append(
                    'Default value of "%s" should have quotes for field %s' % (self._default, self._name)
                )

        if self._default and not (self._default in self.values):
            self._errors.append(
                "Column %s has default value of %s but this is not an allowed value" % (self._name, self._default)
            )

        if self._default is None and not self._null:
            self._warnings.append(
                'Column %s is not null and has no default: you should set a default to avoid MySQL warnings' %
                (self._name)
            )

        # only a few types of field are allowed to use this
        if not self._column_type.lower() in self.allowed_types:
            self._errors.append(
                'Column of type %s is not allowed to have a list of values for column %s' %
                (self._column_type, self._name)
            )



import unittest

from mygrations.formats.mysql.file_reader.create_parser import create_parser
from mygrations.formats.mysql.definitions.constraint import constraint
class test_table_operations_constraints(unittest.TestCase):
    def _get_default_table(self):
        table = create_parser()
        table.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `status_id` int(10) unsigned not null,
            `task` varchar(255) DEFAULT NULL,
            `subject` varchar(255) NOT NULL DEFAULT '',
            PRIMARY KEY (id),
            KEY `tasks_account_id` (`account_id`),
            KEY `tasks_status_id` (`status_id`),
            CONSTRAINT `tasks_account_id_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """
        )

        return table

    def _get_new_constraint(self):
        return constraint(
            **{
                "name": "status_id_fk",
                "column": "status_id",
                "foreign_table": "task_statuses",
                "foreign_column": "id",
                "on_delete": "RESTRICT",
                "on_update": "CASCADE"
            }
        )

    def test_add_constraint(self):
        table = self._get_default_table()
        table.add_constraint(self._get_new_constraint())
        self.assertTrue('status_id_fk' in table.constraints)

    def test_cannot_add_duplicate_constraint(self):
        table = self._get_default_table()
        table.add_constraint(self._get_new_constraint())
        with self.assertRaises(ValueError):
            table.add_constraint(self._get_new_constraint())

    def test_remove_constraint(self):
        table = self._get_default_table()
        table.remove_constraint('tasks_account_id_fk')
        self.assertFalse('tasks_account_id_fk' in table.constraints)

    def test_cannot_remove_missing_constraint(self):
        table = self._get_default_table()
        with self.assertRaises(ValueError):
            table.remove_constraint('asdf')

    def test_change_constraint(self):
        table = self._get_default_table()
        table.change_constraint(
            constraint(
                **{
                    "name": "tasks_account_id_fk",
                    "column": "account_id",
                    "foreign_table": "task_accounts",
                    "foreign_column": "id",
                    "on_delete": "RESTRICT",
                    "on_update": "SET NULL"
                }
            )
        )

        self.assertEquals(
            'CONSTRAINT `tasks_account_id_fk` FOREIGN KEY (`account_id`) REFERENCES `task_accounts` (`id`) ON DELETE RESTRICT ON UPDATE SET NULL',
            str(table.constraints['tasks_account_id_fk'])
        )

    def test_cannot_change_missing_constraint(self):
        table = self._get_default_table()
        with self.assertRaises(ValueError):
            table.change_constraint(self._get_new_constraint())



import importlib

commands = {
    'apply': 'apply',
    'check': 'check',
    'import': 'import_files',
    'plan': 'plan',
    'plan_export': 'plan_export',
    'version': 'version'
    #'export':       'export_files'
}
def allowed(command):

    return command in commands
def execute(command, options):

    if not command in commands:
        raise ValueError("Unknown command %s" % command)

    module = importlib.import_module('.%s' % commands[command], __name__)

    module.execute(options)



import unittest

from mygrations.helpers.dotenv import dotenv, DotEnvSyntaxError
# this does very little really, other than  run .parse_line on each line.
# parse_line is thoroughly tested, so we don't need to do much.
class test_dotenv_parse(unittest.TestCase):

    dotenv = None

    def setUp(self):

        self.dotenv = dotenv()

    def test_json(self):

        # it should detect json and parse it as such
        parsed = self.dotenv.parse('{ "key": "value",\n "name": "bob" }')

        self.assertEquals({'key': 'value', 'name': 'bob'}, parsed)

    def test_dont_interrupt_syntax_error(self):

        # a syntax error from parse_line should not be stopped
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse, 'name')

    def test_parse_each_line(self):

        # otherwise each line should be parsed
        parsed = self.dotenv.parse('key = value\nname = bob #sup')

        self.assertEquals({'key': 'value', 'name': 'bob'}, parsed)

    def test_parse_skip_empty(self):

        # some whitespace checks and a few parts just because
        parsed = self.dotenv.parse('key = value\n\n\n\n   name = "bob" # sup\n another = "test\\"value"#okay')

        self.assertEquals({'key': 'value', 'name': 'bob', 'another': 'test"value'}, parsed)



from mygrations.core.parse.parser import parser
class table_option(parser):

    name = ''
    value = ''

    # UNIQUE account_id (account_id)
    rules = [{
        'type': 'regexp',
        'value': '[^=]{1,30}',
        'name': 'name'
    }, {
        'type': 'literal',
        'value': '='
    }, {
        'type': 'regexp',
        'value': '[^\s;]+',
        'name': 'value'
    }]

    def process(self):

        self._errors = []
        self._warnings = []
        self.name = self._values['name'].strip()
        self.value = self._values['value'].strip()



import unittest

from mygrations.formats.mysql.file_reader.parsers.type_text import type_text
class test_type_text(unittest.TestCase):
    def test_simple(self):

        # parse typical insert values
        parser = type_text()
        returned = parser.parse("name text not null,")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('column', parser.definition_type)
        self.assertEquals('name', parser.name)
        self.assertEquals('TEXT', parser.column_type)
        self.assertFalse(parser.null)
        self.assertTrue(parser.has_comma)
        self.assertEquals('', parser.character_set)
        self.assertEquals('', parser.collate)
        self.assertEquals(0, len(parser.errors))
        self.assertEquals("`name` TEXT NOT NULL", str(parser))

    def test_character_set(self):

        # parse typical insert values
        parser = type_text()
        returned = parser.parse("name text character set 'blah' collate 'boo',")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('column', parser.definition_type)
        self.assertEquals('name', parser.name)
        self.assertEquals('TEXT', parser.column_type)
        self.assertTrue(parser.null)
        self.assertTrue(parser.has_comma)
        self.assertEquals('BLAH', parser.character_set)
        self.assertEquals('BOO', parser.collate)
        self.assertEquals(0, len(parser.errors))
        self.assertEquals("`name` TEXT CHARACTER SET 'BLAH' COLLATE 'BOO'", str(parser))

    def test_strip_backticks(self):

        # parse typical insert values
        parser = type_text()
        returned = parser.parse("`name` text not null")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertEquals('column', parser.definition_type)
        self.assertEquals('name', parser.name)
        self.assertEquals("`name` TEXT NOT NULL", str(parser))

    def test_no_default(self):

        # parse typical insert values
        parser = type_text()
        returned = parser.parse("name text default 'bob'")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertTrue('not allowed to have a default' in parser.errors[0])

    # some types must have a length
    def test_invalid_column_type(self):

        for coltype in ['integer', 'varchar', 'afijeirjeri']:

            parser = type_text()
            returned = parser.parse('name %s' % coltype)

            self.assertTrue(parser.matched)
            self.assertEquals('', returned)
            self.assertTrue('must have a length' in parser.errors[0])

    # only some types can not have a length
    def test_valid_column_type(self):

        for coltype in ['tinyblob', 'blob', 'mediumblob', 'longblob', 'tinytext', 'text', 'mediumtext', 'longtext']:

            parser = type_text()
            returned = parser.parse('name %s' % coltype)

            self.assertTrue(parser.matched)
            self.assertEquals('', returned)
            self.assertEquals(0, len(parser.errors))



class column(object):

    definition_type = 'column'
    _unsigned = None
    _character_set = None
    _collate = None
    _auto_increment = None
    _errors = None
    _warnings = None

    def __init__(
        self,
        name='',
        length='255',
        null=True,
        column_type='VARCHAR',
        default=None,
        unsigned=None,
        character_set='',
        collate='',
        auto_increment=False
    ):
        self._name = name
        self._length = length
        self._null = null
        self._column_type = column_type
        self._default = default
        self._unsigned = unsigned
        self._character_set = character_set
        self._collate = collate
        self._auto_increment = auto_increment

    @property
    def name(self):
        """ Public getter.  Returns the name of the column.

        :returns: The column name
        :rtype: string
        """

        return self._name

    @property
    def length(self):
        """ Public getter.  Returns the length of the column as a string.

        Some examples of the length for various column definitions:

        ==================  ====================
        Type                Length
        ==================  ====================
        INT(10) UNSIGNED    10
        VARCHAR(255)        255
        decimal(20,5)       20,5
        date
        ==================  ====================

        :returns: The column length
        :rtype: string
        """

        return self._length

    @property
    def null(self):
        """ Public getter.  Returns True/False to denote if the column is allowed to be null

        :returns: Whether or not null is an allowed value for the column
        :rtype: bool
        """
        return self._null

    @property
    def column_type(self):
        """ Public getter.  Returns a string denoting the type of the column.  Always returns in uppercase

        Some examples of the length for various column definitions:

        ==================  ====================
        Definition          Type
        ==================  ====================
        INT(10) UNSIGNED    INT
        VARCHAR(255)        VARCHAR
        decimal(20,5)       DECIMAL
        date                DATE
        ==================  ====================

        :returns: The column type
        :rtype: string
        """
        return self._column_type.upper()

    @property
    def default(self):
        """ Public getter.  Returns the default value for the column as a string, or None for a default value of null

        Returns None to represent a default value of null.

        :returns: The default value
        :rtype: string|None
        """
        return self._default

    @property
    def unsigned(self):
        """ Public getter.  Returns True, False, or None to denote the status of the UNSIGNED property

        ==================  ====================
        Return Value        Meaning
        ==================  ====================
        True                The column is unsigned
        False               The column is signed
        None                UNSIGNED is not an applicable property for this column type
        ==================  ====================

        :returns: True, False, or None
        :rtype: bool|None
        """
        return self._unsigned

    @property
    def character_set(self):
        """ Public getter.  Returns None or a value to denote the CHARACTER_SET property

        :returns: string, or None
        :rtype: string|None
        """
        return None if self._character_set is None else self._character_set.upper()

    @property
    def collate(self):
        """ Public getter.  Returns None or a value to denote the COLLATE property

        :returns: string, or None
        :rtype: string|None
        """
        return None if self._collate is None else self._collate.upper()

    @property
    def auto_increment(self):
        """ Public getter.  Returns True, False, or None to denote the status of the AUTO_INCREMENT property

        ==================  ====================
        Return Value        Meaning
        ==================  ====================
        True                The column is an AUTO_INCREMENT column
        False               The column is not an AUTO_INCREMENT column
        None                AUTO_INCREMENT is not an applicable property for this column type
        ==================  ====================

        :returns: True, False, or None
        :rtype: bool|None
        """
        return self._auto_increment

    @property
    def errors(self):
        """ Public getter.  Returns a list of parsing errors

        :returns: A list of parsing errors
        :rtype: list
        """
        return [] if self._errors is None else self._errors

    @property
    def warnings(self):
        """ Public getter.  Returns a list of parsing/table warnings

        :returns: A list of parsing/table warnings
        :rtype: list
        """
        return [] if self._warnings is None else self._warnings

    def __str__(self):
        """ Returns the MySQL command that would create the column

        i.e. column_name type(len) default ''

        :returns: A partial MySQL command that could be used to generate the column
        :rtype: string
        """
        parts = []
        parts.append('`%s`' % self.name)

        type_string = self.column_type
        if self.length:
            if type(self.length) == type([]):
                length = "'%s'" % ("','".join(self.length))
            else:
                length = self.length
            type_string += '(%s)' % length
        parts.append(type_string)

        if self.unsigned:
            parts.append('UNSIGNED')

        if not self.null:
            parts.append('NOT NULL')

        if self.default is not None:
            if self.default == '':
                parts.append("DEFAULT ''")
            elif self.default.isnumeric():
                parts.append("DEFAULT %s" % self.default)
            else:
                parts.append("DEFAULT '%s'" % self.default)

        if self.auto_increment:
            parts.append('AUTO_INCREMENT')

        if self.character_set:
            parts.append("CHARACTER SET '%s'" % self.character_set)

        if self.collate:
            parts.append("COLLATE '%s'" % self.collate)

        return ' '.join(parts)

    def is_really_the_same_as(self, column):
        """ Takes care of a pesky false-positive when checking columns

        :param column: The column to comprehensively check for a difference with
        :type column: mygrations.formats.mysql.definitions.column
        :returns: True if the column really is the same, even for apparent differences
        :rtype: bool
        """
        # if any of these attributes change then it really isn't the same
        for attr in ['name', 'length', 'null', 'column_type', 'unsigned']:
            if getattr(self, attr) != getattr(column, attr):
                return False

        # default needs a special check because it can run into issues for decimal columns
        if self.column_type == 'DECIMAL':
            split = self.length.split(',')
            if len(split) == 2:
                ndecimals = int(split[1])
                if (self.default is None
                    and column.default is not None) or (self.default is not None and column.default is None):
                    return False
                if round(float(self.default), ndecimals) != round(float(column.default), ndecimals):
                    return False
        else:
            if self.default != column.default:
                return False

        # if collate or character_set are different and *both* have a value,
        # then these aren't really the same
        for attr in ['collate', 'character_set']:
            my_val = getattr(self, attr)
            that_val = getattr(column, attr)
            if my_val and that_val and my_val != that_val:
                return False

        # if we got here then these columns are the same. Either all attributes have the same
        # values, or collate and/or character_set differ, but they differ by one not having
        # a value.  This is the false-positive we are trying to check for
        # (see tests.formats.mysql.definitions.test_table_text_false_positives)
        return True



import unittest

from mygrations.core.parse.rule_delimited import rule_delimited
class test_parse_delimited(unittest.TestCase):
    def get_rule(self, name, separator, quote, literal):

        return rule_delimited(
            False, {
                'name': name,
                'separator': separator,
                'quote': quote
            }, {
                'type': 'literal',
                'value': literal
            }
        )

    def test_name_required(self):

        with self.assertRaises(ValueError):

            self.get_rule('', ',', '`', 'asdf')

    def test_separator_required(self):

        with self.assertRaises(ValueError):

            self.get_rule('bob', '', '', 'asdf')

    def test_no_multi_character_separator(self):

        with self.assertRaises(ValueError):

            self.get_rule('bob', 'as', '', 'asdf')

    def test_no_multi_character_quote(self):

        with self.assertRaises(ValueError):

            self.get_rule('bob', ',', 'as', 'asdf')

    def test_literal_required(self):

        with self.assertRaises(ValueError):

            rule_delimited(False, {'name': 'bob', 'separator': ',', 'quote': '`'}, {})

    def test_can_init_with_name_and_separator(self):

        rule = self.get_rule('bob', ',', '', 'asdf')
        self.assertEquals(rule.name, 'bob')
        self.assertEquals(rule.separator, ',')

    def test_parse_without_quote(self):

        rule = self.get_rule('bob', ',', '', ')')
        self.assertTrue(rule.parse('1,2,3,4)'))
        self.assertEquals(['1', '2', '3', '4'], rule.result)
        self.assertEquals(')', rule.leftovers)

    def test_parse_optional_quotes(self):

        rule = self.get_rule('bob', ',', '`', ')')
        self.assertTrue(rule.parse('asdf,`bob`,huh,`okay`) sup'))
        self.assertEquals(['asdf', 'bob', 'huh', 'okay'], rule.result)
        self.assertEquals(') sup', rule.leftovers)

    def test_syntax_error_missing_quote(self):

        with self.assertRaises(SyntaxError):

            rule = self.get_rule('bob', ',', '`', ')')
            rule.parse('asdf,`bob)')

    def test_separator_in_quotes(self):

        rule = self.get_rule('bob', ',', '`', ')')
        self.assertTrue(rule.parse('asdf,`bob,`,huh,`okay`) sup'))
        self.assertEquals(['asdf', 'bob,', 'huh', 'okay'], rule.result)
        self.assertEquals(') sup', rule.leftovers)

    def test_alternate_characters(self):

        rule = self.get_rule('bob', 'X', '<', 'asdf')

        self.assertTrue(rule.parse('<hey<X<sup<asdf'))
        self.assertEquals(['hey', 'sup'], rule.result)
        self.assertEquals('asdf', rule.leftovers)



from mygrations.core.parse.parser import parser
from mygrations.formats.mysql.definitions.rows import rows

from .parsers.insert_values import insert_values
class insert_parser(parser, rows):

    rules = [{
        'type': 'literal',
        'value': 'INSERT INTO'
    }, {
        'type': 'regexp',
        'value': '[^\s\(]+',
        'name': 'table'
    }, {
        'type': 'literal',
        'value': '('
    }, {
        'type': 'delimited',
        'name': 'columns',
        'separator': ',',
        'quote': '`'
    }, {
        'type': 'literal',
        'value': ')'
    }, {
        'type': 'literal',
        'value': 'VALUES'
    }, {
        'type': 'children',
        'name': 'inserts',
        'classes': [insert_values]
    }, {
        'type': 'literal',
        'value': ';',
        'optional': True,
        'name': 'closing_semicolon'
    }]

    def __init__(self, rules=[]):

        super().__init__(rules)

        self._columns = []
        self._raw_rows = []

    def process(self):

        self._table = self._values['table'].strip('`')
        self._columns = self._values['columns']
        self._raw_rows = []
        self._num_explicit_columns = len(self._columns) if self._columns else 0
        self._errors = []
        self._warnings = []
        self.has_semicolon = True if 'closing_semicolon' in self._values else False

        had_comma = True
        for row in self._values['inserts']:

            if not had_comma:
                self._warnings.append('Missing comma between insert value groups')

            if self._num_explicit_columns and len(row.values) != self._num_explicit_columns:
                self._errors.append('Insert values has wrong number of values for %s' % (row))
                continue

            self._raw_rows.append(row.values)
            had_comma = True if row.has_comma else False



from .rule_base import rule_base
class rule_literal(rule_base):
    """ rule = rule_children( parser, rule, next_rule )

    rule_literal constructor.  Pass in the parser object that the rule is for, the rule
    that is being parsed, and the next rule in the parsing chain.

    This class attempts to match a literal string (case-insensitive).

    keys for the rules dictionary:

    =========  ========
    key        contents
    =========  ========
    value      The literal value to look for
    name       (optional) The name of the rule
    optional   (optional) If true, denotes that this rule is optional
    =========  ========

    :param parser: The parser that this rule belongs to
    :param rule: A dictionary representing the rule configuration (see each rule class for details)
    :param next_rule: A dictionary representing the next rule
    :type parser: parser
    :type rule: dict
    :type next_rule: dict
    :returns: rule object
    """

    # names are not required for literals
    require_name = False

    literal = ''

    def __init__(self, parser, rule, next_rule):

        super().__init__(parser, rule, next_rule)

        self.literal = self.rule['value']

        # use the actual literal value as the name if we don't have one
        if not self.name:
            self.name = self.literal

    def parse(self, string):

        # easy to check for a literal match at the beginning of string
        val_len = len(self.literal)
        if string[:val_len].lower() != self.literal.lower():
            self.result = ''
            self.leftovers = string
            return False

        # if we matched then clean!
        self.leftovers = string[val_len:].strip()
        self.result = self.literal
        return True



import os, os.path
from mygrations.helpers.dotenv import dotenv
from mygrations.helpers.db_credentials import db_credentials
class base(object):

    credentials = {}
    config = {}

    def __init__(self, options):

        self.options = options

        if not 'env' in self.options:
            raise ValueError('Missing "env" in options for commands.import_files')

        if not 'config' in self.options:
            raise ValueError('Missing "config" in options for commands.import_files')

        # load up the mygration configuration (which includes the path to the files we will import)
        config_abs_path = self._resolve_and_check_config_file(self.options['config'])
        self.config = dotenv(config_abs_path)
        self.config['files_directory'] = self._relative_to_config_abs_path(
            self.config['files_directory'],
            config_abs_path
        )
        self.options['env'] = self._relative_to_config_abs_path(
            self.options['env'],
            config_abs_path
        )

        # and load up the database credentials
        self.credentials = db_credentials(self.options['env'], self.config)

        if not 'files_directory' in self.config:
            raise ValueError('Missing files_directory configuration setting in configuration file')

    def execute(self):
        raise NotImplementedError()

    def _resolve_and_check_config_file(self, file_path):
        if not len(file_path):
            raise ValueError('Missing path to env file')

        # We want to find the location of the mygrate.conf file.  For the most part expect file_path to literally
        # be `mygrate.conf`, as that is the default sent in by the runner.  However the user could have specified
        # something else.  If we have an absolute path specified then that is the only thing we want to check,
        # and we want to throw an exception if it does not match.  However, if it is not an absolute path we
        # want to search for the given file in current directory or its parents.
        if file_path[0] == '/':
            if not os.path.isfile(file_path):
                raise ValueError("Specified env file '%s' does not exist" % file_path)
            return file_path

        directories = ('%s/%s' % (os.getcwd(), file_path)).strip('/').split('/')
        filename = directories.pop()
        ndirectories = len(directories)
        for i in range(ndirectories):
            abs_path = '/%s/%s' % ('/'.join(directories[:ndirectories-i]), filename)
            if os.path.isfile(abs_path):
                return abs_path

        raise ValueError("Could not find env file '%s' in current directory or parents" % file_path)

    def _relative_to_config_abs_path(self, relative_path, config_abs_path):
        # Our file directory and our .env file path should both be relative to the location of the mygrate.conf file.
        # The exact path to the mygrate.conf file can change because it is allowed to be in a parent of the
        # current working directory.  Therefore we need a way to take a relative path and turn it into an absolute
        # path using the absolute path to the configuration file (i.e. the mygrate.conf file).
        # This method does just that.
        # However, don't muck with a path if it already absolute
        if relative_path[0] == '/':
            return relative_path

        return '%s/%s' % (os.path.dirname(config_abs_path), relative_path)



import unittest

from mygrations.formats.mysql.db_reader.database import database as database_reader
from mygrations.drivers.mysqldb.mysqldb import mysqldb
from tests.mocks.db.mysql.db_structure import db_structure
class test_database(unittest.TestCase):
    def _get_tables(self):

        return {
            'logs':
            """
                CREATE TABLE `logs` (
                    `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                    `message` TEXT NOT NULL,
                    `traceback` text,
                    PRIMARY KEY (`id`)
                ) ENGINE=InnoDB DEFAULT CHARSET=utf8;""",
            'more_logs':
            """
                CREATE TABLE `more_logs` (
                    `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                    `more_messages` TEXT NOT NULL,
                    `traceback` text,
                    PRIMARY KEY (`id`)
                ) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        }

    def test_simple(self):

        tables = self._get_tables()

        mock_db = db_structure(tables, {})
        database = database_reader(mysqldb(mock_db))

        # our parser should have a table!
        self.assertTrue('logs' in database.tables)
        self.assertTrue('more_logs' in database.tables)

    def test_read_rows(self):

        tables = self._get_tables()
        rows = {
            'logs': ({
                'id': 1,
                'message': 'hey',
                'traceback': 'never'
            }, {
                'id': 2,
                'message': 'sup',
                'traceback': 'forever'
            })
        }

        database = database_reader(mysqldb(db_structure(tables, rows)))

        # quick double check
        self.assertTrue('logs' in database.tables)
        self.assertTrue('more_logs' in database.tables)

        database.read_rows('logs')

        self.assertTrue(database.tables['logs'].tracking_rows)
        rows = database.tables['logs'].rows
        self.assertEquals(1, rows[1]['id'])
        self.assertEquals('hey', rows[1]['message'])
        self.assertEquals('never', rows[1]['traceback'])

        self.assertEquals(2, rows[2]['id'])
        self.assertEquals('sup', rows[2]['message'])
        self.assertEquals('forever', rows[2]['traceback'])



from .dotenv import dotenv
class db_credentials(dict):

    config = {}
    env = {}

    key_map = {'hostname_key': 'host', 'database_key': 'db', 'username_key': 'user', 'password_key': 'passwd'}

    def __init__(self, env, config):

        # Config represents the contents of the mygrate.conf file.
        # env represents the contents of the applications .env file.
        # Normally the config file will tell us which keys in the .env
        # file have the database credentials.
        self.config = config if isinstance(config, dotenv) else dotenv(config)
        self.env = env if isinstance(env, dotenv) else dotenv(env)

        # we need four pieces of information for this to work:
        for key in ['hostname_key', 'username_key', 'password_key', 'database_key']:
            if not key in self.config:
                raise ValueError("Cannot find required key %s in mygrate.conf" % key)

        # now make sure we have it in our .env
        credentials = {}
        for (key, val) in self.key_map.items():

            key_name = self.config[key]
            if not key_name in self.env:
                raise ValueError("Failed to fulfill %s: key %s not found in .env" % (val, key_name))

            credentials[val] = self.env[key_name]

        super().__init__(credentials)



class database(object):

    _errors = None
    _warnings = None
    _tables = None
    _rows = None
    _errors_1215 = None

    def __init__(self):
        self._warnings = []
        self._errors = []
        self._tables = {}
        self._rows = []
        self._errors_1215 = None

    @property
    def tables(self):
        """ Public getter.  Returns a dict of table definitions, by table name

        :returns: A dict of table definitions, by table name
        :rtype: dict
        """
        return self._tables

    @property
    def errors(self):
        """ Public getter.  Returns a list of parsing errors

        :returns: A list of parsing errors
        :rtype: list
        """
        return self._errors

    @property
    def warnings(self):
        """ Public getter.  Returns a list of parsing/table warnings

        :returns: A list of parsing/table warnings
        :rtype: list
        """
        return self._warnings

    @property
    def errors_1215(self):
        """ Public getter.  Returns a list of 1215 errors found for the database

        :returns: A list of MySQL 1215 error messages
        :rtype: [string]
        """
        if self._errors_1215 is None:
            self._errors_1215 = self._find_all_1215_errors()
        return self._errors_1215

    def store_rows_with_tables(self):
        """ Processes table rows and adds them to the appropriate tables

        Table rows are stored with tables for comparison purposes, but might
        come in through their own separate files.  This method puts the two
        together.
        """
        for rows in self._rows:
            if not rows.table in self._tables:
                self._errors.append('Found rows for table %s but that table does not have a definition' % rows.table)
                continue

            returned = self._tables[rows.table].add_rows(rows)
            if isinstance(returned, str):
                self._errors.append(returned)

    def unfulfilled_fks(self, table):
        """ Returns a dictionary with information about all constraints in the table which are not fulfilled by this database

        If all foreign keys are fulfilled by the database structure then an empty dict is returned

        The returned dictionary contains a key with the name of every foreign key that cannot
        be fulfilled.  The value in the dictionary will be another dictionary containing
        'error' (an error message stating exactly what the problem is) and 'foreign_key'
        (the actual foreign key definition that cannot be fulfilled)

        :param table: The table to check
        :type table: mygrations.formats.mysql.definitions.table
        :return: Dictionary with information on all foreign keys that cannot be fulfilled
        :rtype: dict
        """
        if not table.constraints:
            return {}

        unfulfilled = {}
        for (constraint_name, constraint) in table.constraints.items():
            error = self.find_1215_errors(table, constraint)
            if error:
                unfulfilled[constraint_name] = {"error": error, "foreign_key": constraint}

        return unfulfilled

    def _find_all_1215_errors(self):
        """ Returns a list of any 1215 error messages for this database

        :return: A list of 1215 error messages
        :rtype: [string]
        """
        errors = []
        for table in self.tables.values():
            if not table.constraints:
                continue

            for constraint in table.constraints.values():
                error = self.find_1215_errors(table, constraint)
                if error:
                    errors.append(error)

        return errors

    def find_1215_errors(self, table, constraint):
        """ Returns None or a string describing a 1215 error message found for the given table and constraint

        :param table: The table being checked
        :param constraint: The constraint to check the table against
        :type table: mygrations.formats.mysql.definitions.table
        :type constraint: mygrations.formats.mysql.definitions.constraint
        :rtype: string|None
        """

        if constraint.foreign_table not in self.tables:
            return "MySQL 1215 error for foreign key `%s`: `%s`.`%s` references `%s`.`%s`, but table `%s` does not exist" % (
                constraint.name, table.name, constraint.column, constraint.foreign_table, constraint.foreign_column,
                constraint.foreign_table
            )
        foreign_table = self.tables[constraint.foreign_table]
        if constraint.foreign_column not in foreign_table.columns:
            return "MySQL 1215 error for foreign key `%s`: `%s`.`%s` references `%s`.`%s`, but column `%s`.`%s` does not exist" % (
                constraint.name, table.name, constraint.column, constraint.foreign_table, constraint.foreign_column,
                constraint.foreign_table, constraint.foreign_column
            )

        # the column exists but we may still have a 1215 error.  That can happen in a few ways
        table_column = table.columns[constraint.column]
        foreign_column = foreign_table.columns[constraint.foreign_column]

        if not table.column_is_indexed(table.columns[table_column.name]):
            return "MySQL 1215 error for foreign key `%s`: missing index. `%s`.`%s` does not have an index and therefore cannot be used in a foreign key constraint" % (
                constraint.name, table.name, table_column.name
            )

        # we have a few attributes that must must match exactly and have easy-to-produce errors
        for attr in ['column_type', 'length', 'character_set', 'collate']:
            table_value = getattr(table_column, attr)
            foreign_value = getattr(foreign_column, attr)
            if table_value != foreign_value:
                return "MySQL 1215 error for foreign key `%s`: %s mismatch. `%s`.`%s` is '%s' but `%s`.`%s` is '%s'" % (
                    constraint.name, attr.replace('_', ' '), table.name, constraint.column, table_value,
                    foreign_table.name, foreign_column.name, foreign_value
                )

        # unsigned are separate because they get a slightly different message
        if table_column.unsigned and not foreign_column.unsigned:
            return "MySQL 1215 error for foreign key `%s`: unsigned mistmatch. `%s`.`%s` is unsigned but `%s`.`%s` is not" % (
                constraint.name, table.name, table_column.name, foreign_table.name, foreign_column.name
            )

        if not table_column.unsigned and foreign_column.unsigned:
            return "MySQL 1215 error for foreign key `%s`: unsigned mistmatch. `%s`.`%s` is unsigned but `%s`.`%s` is not" % (
                constraint.name, foreign_table.name, foreign_column.name, table.name, table_column.name
            )

        # if the constraint has a SET NULL but the column cannot be null, then 1215
        if (constraint.on_delete == 'SET NULL' or constraint.on_update == 'SET NULL') and not table_column.null:
            message_parts = []
            if constraint.on_delete == 'SET NULL':
                message_parts.append('ON DELETE')
            if constraint.on_update == 'SET NULL':
                message_parts.append('ON UPDATE')
            return "MySQL 1215 error for foreign key `%s`: invalid SET NULL. `%s`.`%s` is not allowed to be null but the foreign key attempts to set the value to null %s" % (
                constraint.name, table.name, table_column.name, ' and '.join(message_parts)
            )

        # if the column the constraint is on doesn't have an index, then 1215
        if not foreign_table.column_is_indexed(foreign_column):
            return "MySQL 1215 error for foreign key `%s`: missing index. `%s`.`%s` references `%s`.`%s` but `%s`.`%s` does not have an index and therefore cannot be used in a foreign key constraint" % (
                constraint.name, table.name, table_column.name, foreign_table.name, foreign_column.name,
                foreign_table.name, foreign_column.name
            )

        return None

    def add_table(self, table):
        """ Adds a table to the database

        :param table: The table to add
        :type table: mygrations.formats.mysql.definitions.table
        """
        if table.name in self._tables:
            raise ValueError('Cannot add table %s to database because it already exists' % table.name)

        self._errors_1215 = None
        self._tables[table.name] = table

    def remove_table(self, table):
        """ Removes a table from the database

        :param table: The table to remove
        :type table: mygrations.formats.mysql.definitions.table
        """
        if not table.name in self._tables:
            raise ValueError('Cannot remove table %s from database because it does not exist' % table.name)

        self._errors_1215 = None
        self._tables.pop(table.name, None)

    def apply_operation(self, table_name, operation):
        """ Updates the database object according to the operation

        :param table_name: The table that the operation is being applied to
        :param operation: The operation to apply
        :type table: string|mygrations.formats.mysql.definitions.table
        :type operation: mygrations.formats.mysql.mygration.operations.*
        """
        if type(table_name) != str:
            table_name = table_name.name

        if not table_name in self._tables:
            raise ValueError('Cannot apply operation to table %s because that table does not exist' % table_name)

        # the table applies the operation
        self._errors_1215 = None
        self._tables[table_name].apply_operation(operation)

    def apply_to_source(self, operation):
        """ Updates the actual source of the database object: usually the database or a .sql file

        This method and :meth:`mygrations.formats.mysql.defintions.database.apply_operation` sound very similar.
        The difference is that the latter operates on the in-memory structure store only, and makes
        no actual changes.  Instead it is primarily used for internal purposes.  This method actually
        makes changes.

        :param operation: The operation to apply
        :type operation: mygrations.formats.mysql.mygration.operations.*
        """
        raise NotImplementedError("Someone forgot to finish something")



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='mygrations',
    version='0.13.0',
    description='A general purpose migration tool for managing MySQL updates',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/cmancone/mygrations',
    author='Conor Mancone',
    author_email='cmancone@gmail.com',
    license='MIT',
    download_url='https://github.com/cmancone/mygrations/archive/v0.13.tar.gz',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    keywords='setuptools development migrations mysql',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    # data_files=[('/usr/local/bin', ['mygrate.py'])],
)



from mygrations.core.parse.parser import parser
class comment_parser(parser):

    rules = ['nope']
    sql = ''

    def __init__(self, rules=[]):

        # we're not yet a rules-based parser (I may do that eventually), so
        # I don't want to run the normal parser __init__()
        self.errors = []
        self.warnings = []
        self.matched = False

    def parse(self, sql):
        """ res = paser.parse()

        Parses the SQL, storing the comment and returning an SQL string with the comment removed

        :param sql: The SQL string to parse
        :type sql: string
        :returns: The SQL string with the first comment removed
        :rtype: string

        :Example:
                >>> import comment_parser
                >>> parser = comment_parser( '--Some SQL\nINSERT INTO some_table (1,2)' )
                >>> without_comment = parser.parse()
                >>> print( parser.comment )
                Some SQL
                >>> print( without_comment )
                INSERT INTO some_table (1,2)
        """

        self.sql = sql.strip()
        self.comment = ''

        # what kind of comment are we dealing with?
        # -- and # go to the end of the line.  /* goes to */
        if self.sql[0] == '#' or self.sql[:2] == '--':

            # which is really easy if there is no newline: the
            # whole thing is our comment and there is no data left
            if not self.sql.count('\n'):
                self.comment = self.sql
                self._values = {'commment': self.comment}
                self.matched = True
                return ''

            # otherwise just find the newline and return the rest
            else:
                index = self.sql.index('\n')
                self.comment = self.sql[:index]
                self._values = {'commment': self.comment}
                self.matched = True
                return self.sql[index + 1:].strip()

        # then we should be dealing with /* ... */.  Our line should
        # start with it or we have a problem
        if self.sql[:2] != '/*':
            raise ValueError('SQL passed to comment parser did not start with a comment')

        if not self.sql.count('*/'):
            self.errors.append('Could not find closing comment indicator, */')
            return self.sql

        # otherwise this is very straight-forward
        index = self.sql.index('*/')
        self.comment = self.sql[2:index].strip()
        self._values = {'commment': self.comment}
        self.matched = True
        return self.sql[index + 2:].strip()



import unittest

from mygrations.formats.mysql.file_reader.reader import reader
from mygrations.formats.mysql.file_reader.database import database
class test_regressions(unittest.TestCase):
    """ Some tests based off failing CREATE TABLE commands from our database when things were first being finished """

    def test_payment_request_types(self):

        parser = reader()
        returned = parser.parse(
            """CREATE TABLE `payment_request_types` (
  `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
  `name` VARCHAR(255) COLLATE utf8_unicode_ci NOT NULL DEFAULT '',
  `slug` VARCHAR(255) COLLATE utf8_unicode_ci NOT NULL DEFAULT '',
  `order_by` INT(11) NOT NULL DEFAULT 0,
  PRIMARY KEY (`id`),
  KEY `order_by_payment_request_types` (`order_by`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_unicode_ci;

INSERT INTO payment_request_types (id,name,slug,order_by) VALUES (1,'Standard','standard',1),(2,'Special Circumstances','emergency',2);"""
        )

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have matched everything
        self.assertEquals('', returned)

        # we should not have any errors
        self.assertEquals([], parser.errors)

    def test_employee_contract_list(self):

        db = database(
            """CREATE TABLE `employee_contract_list` (
 `id` int(10) NOT NULL AUTO_INCREMENT,
 `account_id` int(10) unsigned NOT NULL,
 `name` varchar(255) NOT NULL,
 `display` varchar(255) NOT NULL,
 `table_name` varchar(255) NOT NULL,
 `employee_field_name` varchar(255) NOT NULL,
 `link` varchar(255) NOT NULL,
 PRIMARY KEY (`id`),
 KEY `employee_contact_list_account_id_ref_accounts_id` (`account_id`),
 CONSTRAINT `employee_contact_list_account_id_ref_accounts_id` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE
) ENGINE=InnoDB DEFAULT CHARSET=utf8;

INSERT INTO `employee_contract_list` (`id`, `account_id`, `name`, `display`, `table_name`, `employee_field_name`, `link`) VALUES
(1, 1, 'main_contract', 'Employee Contract', 'employee_main_contracts', 'main_contract_id', 'hr/contracts/sign_main_contract'),
(2, 1, 'nda', 'Non-Disclosure Agreement ', 'employee_ndas', 'nda_id', 'hr/contracts/sign_nda'),
(3, 1, 'non_compete', 'Employee No Compete Agreement', 'employee_no_competes', 'no_compete_id', 'hr/contracts/complete_non_compete'),
(4, 1, 'dd_auth', 'Direct Deposit Authorization', 'employee_dd_auths', 'dd_auth_id', 'hr/contracts/complete_dd_auth'),
(5, 1, 'property_agreement', 'Company Property Return/Receipt', 'employee_property_agrees', 'property_receipt_id', 'hr/contracts/property_agreement'),
(6, 1, 'background_ck_agreement', 'Background Check Agreement', 'employee_background_ck_releases', 'background_ck_id', 'hr/contracts/background_ck_agreement'),
(7, 1, 'employee_handbook', 'Employee Handbook Receipt', 'employee_handbook_receipts', 'handbook_receipt_id', 'hr/contracts/sign_employee_handbook'),
(8, 1, 'cell_phone_policy', 'Company Cell Phone Policy', 'employee_cell_policy_receipts', 'cell_policy_receipt_id', 'hr/contracts/sign_cell_phone_policy'),
(9, 1, 'fed_w4', 'Federal W-4', 'employee_fed_w4s', 'fed_w4_id', 'hr/tax_forms/complete_fed_w4'),
(10, 1, 'ga_w4', 'Georgia W-4', 'employee_ga_w4s', 'ga_w4_id', 'hr/tax_forms/complete_ga_w4'),
(11, 1, 'al_a4', 'Alabama A-4', 'employee_al_a4s', 'al_w4_id', 'hr/tax_forms/complete_al_a4'),
(12, 1, 'fed_i9', 'Federal I-9', 'employee_fed_i9s', 'fed_i9_id', 'hr/tax_forms/complete_fed_i9');"""
        )

        self.assertEquals([], db.errors)



import unittest

from mygrations.formats.mysql.file_reader.database import database as database_reader
class test_database(unittest.TestCase):
    def test_simple(self):

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """, """
            CREATE TABLE `more_logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `more_messages` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        database = database_reader(strings)

        # our parser should have a table!
        self.assertTrue('logs' in database.tables)
        self.assertTrue('more_logs' in database.tables)

    def test_with_rows(self):

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """, """
            INSERT INTO logs (id,message,traceback) VALUES (1,'hi','sup'),(2,'what','kay');
        """
        ]
        database = database_reader(strings)

        # our parser should have a table!
        self.assertTrue('logs' in database.tables)
        self.assertEquals(3, database.tables['logs'].auto_increment)

        rows = database.tables['logs'].rows
        self.assertEquals('hi', rows[1]['message'])
        self.assertEquals('sup', rows[1]['traceback'])
        self.assertEquals('what', rows[2]['message'])
        self.assertEquals('kay', rows[2]['traceback'])

    def test_keeps_errors(self):

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT DEFAULT '',
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        database = database_reader(strings)

        self.assertTrue('not allowed to have a default value for column message in table logs' in database.errors[0])



class remove_table:
    """ Generates an SQL command to drop a table """

    def __init__(self, table_name):
        if type(table_name) != str:
            self._table_name = table_name.name
        else:
            self._table_name = table_name

    @property
    def table_name(self):
        """ Public getter.  Returns the name of the table.

        :returns: The table name
        :rtype: string
        """

        return self._table_name

    def __str__(self):
        return 'DROP TABLE %s;' % self._table_name



import unittest

from mygrations.helpers.dotenv import dotenv, DotEnvSyntaxError
class test_dotenv_parse_line(unittest.TestCase):

    dotenv = None

    def setUp(self):

        self.dotenv = dotenv()

    def test_opening_comment(self):

        self.assertEquals([False, False], self.dotenv.parse_line('# name = test', '#'))

    def test_opening_comment_with_space(self):

        # white space is stripped first, so it still starts with a comment if it has spaces
        self.assertEquals([False, False], self.dotenv.parse_line(' # name = test', '#'))

    def test_any_comment_charater(self):

        # our comment character should be respected
        self.assertEquals([False, False], self.dotenv.parse_line('; name = test', ';'))

    def test_multi_character_comment(self):

        # and it is allowed to be more than one character long
        self.assertEquals([False, False], self.dotenv.parse_line('weirdcomment name = test', 'weirdcomment'))

    def test_ignore_empty_lines(self):

        # empty lines (which includes lines with all white space) are ignored
        self.assertEquals([False, False], self.dotenv.parse_line('  ', '#'))

    def test_invalid_format(self):

        # supported format is something like name = value or name: value
        # if neither of the separators is present then we should get a syntax error
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse_line, 'sup', '#')

    def test_empty_value_equal(self):

        # empty values are allowed
        self.assertEquals(['key', ''], self.dotenv.parse_line('key=', '#'))

    def test_empty_value_colon(self):

        # colons act exactly the same way
        self.assertEquals(['key', ''], self.dotenv.parse_line('key:', '#'))

    def test_empty_value_equal_spaces(self):

        # and spaces are ignored
        self.assertEquals(['key', ''], self.dotenv.parse_line(" key \t=\t  ", '#'))

    def test_empty_value_equal_colon(self):

        # and spaces are ignored
        self.assertEquals(['key', ''], self.dotenv.parse_line(" key \t:\t  ", '#'))

    def test_value_no_quotes_equal(self):

        # a value without quotes should be easy
        self.assertEquals(['key', 'asdf'], self.dotenv.parse_line("key=asdf", '#'))

    def test_value_no_quotes_colon(self):

        # a value without quotes should be easy
        self.assertEquals(['key', 'asdf'], self.dotenv.parse_line("key:asdf", '#'))

    def test_value_no_quotes_equal_spaces(self):

        # spaces are still ignored at the beginning/end of a part
        self.assertEquals(['key', 'asdf bob'], self.dotenv.parse_line("key= asdf bob  \t", '#'))

    def test_value_no_quotes_equal_colon(self):

        # spaces are still ignored at the beginning/end of a part
        self.assertEquals(['key', 'asdf bob'], self.dotenv.parse_line("key : asdf bob  \t", '#'))

    def test_value_no_quotes_equal_comment(self):

        # and comments at the end are ignored (including spaces before comments)
        self.assertEquals(['key', 'asdf bob'], self.dotenv.parse_line("key = asdf bob  \t# a comment", '#'))

    def test_value_no_quotes_equal_colon(self):

        # and comments at the end are ignored (including spaces before comments)
        self.assertEquals(['key', 'asdf bob'], self.dotenv.parse_line("key : asdf bob  \t; a comment", ';'))



from .base import base
import mygrations
def execute(options):

    obj = version(options)
    obj.execute()
class version(base):
    def execute(self):
        print('mygrations v%s' % mygrations.__version__)
        print('Copyright (c) 2017 Conor Mancone (cmancone@gmail.com)')
        print('MIT License')



import unittest

from mygrations.formats.mysql.file_reader.database import database as database_reader
from mygrations.formats.mysql.file_reader.create_parser import create_parser
class test_database(unittest.TestCase):
    def _get_sample_db(self):

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """, """
            CREATE TABLE `more_logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `more_messages` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        return database_reader(strings)

    def test_simple(self):

        db1 = self._get_sample_db()
        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """, """
            CREATE TABLE `less_logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `more_messages` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        db2 = database_reader(strings)

        #differences = db2 - db1
        #self.assertEquals( [], differences )

    def test_add_table(self):
        db = self._get_sample_db()

        new_table = create_parser()
        new_table.parse(
            """CREATE TABLE `log_changes` (
            `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `log_id` INT(10) UNSIGNED NOT NULL,
            `type_id` INT(10) UNSIGNED NOT NULL,
            `change` VARCHAR(255),
            PRIMARY KEY (id),
            KEY `log_changes_log_id` (`log_id`),
            KEY `log_changes_type_id` (`type_id`)
            );
        """
        )

        db.add_table(new_table)

        self.assertEquals(3, len(db.tables))
        self.assertTrue('log_changes' in db.tables)
        self.assertEquals(new_table, db.tables['log_changes'])

    def test_remove_table(self):

        db1 = self._get_sample_db()
        db1.remove_table(db1.tables['more_logs'])

        self.assertEquals(1, len(db1.tables))
        self.assertTrue('logs' in db1.tables)
        self.assertFalse('more_logs' in db1.tables)

    def test_exception_on_remove_invalid_table(self):

        db1 = self._get_sample_db()

        new_table = create_parser()
        new_table.parse(
            """CREATE TABLE `log_changes` (
            `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `log_id` INT(10) UNSIGNED NOT NULL,
            `type_id` INT(10) UNSIGNED NOT NULL,
            `change` VARCHAR(255),
            PRIMARY KEY (id),
            KEY `log_changes_log_id` (`log_id`),
            KEY `log_changes_type_id` (`type_id`)
            );
        """
        )

        with self.assertRaises(ValueError):
            db1.remove_table(new_table)



from collections import OrderedDict

from mygrations.core.parse.parser import parser
from mygrations.formats.mysql.definitions.constraint import constraint
from mygrations.formats.mysql.definitions.index import index
from mygrations.formats.mysql.definitions.column import column
from mygrations.formats.mysql.definitions.table import table

from .parsers import *
class create_parser(parser, table):

    # this defines the rules for the parsing engine.  Yes, I decided to try to build
    # a parsing engine to parse the MySQL.  Seems like a reasonable choice at the
    # moment, but we'll see how it works out in the long run.  The setup you see here
    # is a compromise between the declarative programming principles I like and the
    # time it takes to actually build such a system, especially for a relatively
    # limited use-case.  In the long run if this works well it could probably
    # become its own multi-purpose tool.  These "rules" basically boil down to:
    #
    # CREATE TABLE table_name ( [definitions] ) [options];
    #
    # definitions and options have their own patterns of allowed values and rules,
    # which also have to be defined.  That is handled below for the sake of brevity
    rules = [{
        'type': 'literal',
        'value': 'CREATE TABLE'
    }, {
        'type': 'regexp',
        'value': '\S+',
        'name': 'name'
    }, {
        'type': 'literal',
        'value': '('
    },
             {
                 'type':
                 'children',
                 'name':
                 'definitions',
                 'classes': [
                     index_primary, index_key, index_unique, constraint_foreign, type_character, type_numeric,
                     type_decimal, type_text, type_enum, type_plain
                 ]
             }, {
                 'type': 'literal',
                 'value': ')'
             }, {
                 'type': 'children',
                 'name': 'table_options',
                 'classes': [table_option],
                 'optional': True
             }, {
                 'type': 'literal',
                 'value': ';',
                 'optional': True,
                 'name': 'closing_semicolon'
             }]

    def process(self):

        self.semicolon = True if 'closing_semicolon' in self._values else False
        self._name = self._values['name'].strip('`')
        self._definitions = self._values['definitions']
        self._options = self._values['table_options'] if 'table_options' in self._values else []
        self._columns = OrderedDict()
        self._indexes = OrderedDict()
        self._constraints = OrderedDict()
        self._errors = []
        self._warnings = []
        self._primary = ''

        # ignore the AUTO_INCREMENT option: there is no reason for us to ever manage that
        self._options = [opt for opt in self._options if opt.name != 'AUTO_INCREMENT']

        for definition in self._definitions:
            if isinstance(definition, column):
                self._columns[definition.name] = definition
            elif isinstance(definition, index):
                self._indexes[definition.name] = definition

                if definition.index_type == 'PRIMARY':
                    if self._primary:
                        self._errors.append('Found more than one primary column for table %s' % (self._name))
                    else:
                        self._primary = definition
            elif isinstance(definition, constraint):
                self._constraints[definition.name] = definition

            if definition.errors:
                for error in definition.errors:
                    self._errors.append('%s in table %s' % (error, self._name))

            if definition.warnings:
                for warning in definition.warnings:
                    self._warnings.append('%s in table %s' % (warning, self._name))

        if not self._name:
            self._errors.append('Table name is required')

        if not len(self._columns):
            self._errors.append("Table %s has no columns" % self._name)

        if not self.semicolon:
            self._errors.append("Missing ending semicolon for table %s" % self._name)



__ver__ = '0.1'

from .core import commands
from .helpers.dotenv import dotenv
class mygrate(object):

    command = ''
    options = {}
    config = {}

    def __init__(self, command, options={}):

        # store the name of the command we are executing, and any options that go with it
        self.command = command if command else 'check'

        self.options = options
        if self.options['version']:
            self.command = 'version'

        # make sure we have an actual command
        if not commands.allowed(self.command):
            raise ValueError("Unknown command %s" % self.command)

    def execute(self):

        commands.execute(self.command, self.options)



from .base import base
from mygrations.formats.mysql.file_reader.database import database as database_parser
from mygrations.formats.mysql.db_reader.database import database as database_reader
from mygrations.formats.mysql.mygrations.mygration import mygration
from mygrations.formats.mysql.mygrations.row_mygration import row_mygration
from mygrations.drivers.mysqldb.mysqldb import mysqldb
from mygrations.formats.mysql.mygrations.operations.disable_checks import disable_checks
from mygrations.formats.mysql.mygrations.operations.enable_checks import enable_checks
import sys
def execute(options):

    obj = plan(options)
    obj.execute()
class plan(base):
    def execute(self):

        commands = self.build_commands()
        print(''.join(['%s\n' % command for command in commands]))

    def build_commands(self):

        files_database = database_parser(self.config['files_directory'])

        # any errors or warnings?
        quit_early = False
        if files_database.errors and not self.options['force']:
            print('Errors found in *.sql files', file=sys.stderr)
            quit_early = True
            for error in files_database.errors:
                print(error, file=sys.stderr)

        # or 1215 errors?
        if files_database.errors_1215 and not self.options['force']:
            print('1215 errors encountered', sys.stderr)
            quit_early = True
            for error in files_database.errors_1215:
                print(error, file=sys.stderr)

        if quit_early:
            return []

        # use the credentials to load up a database connection
        live_database = database_reader(mysqldb(self.credentials))

        # we have to tell the live database to load records
        # for any tables we are tracking records for.
        # We use the "files" database as the "truth" because
        # just about any database can have records, but
        # that doesn't mean we want to track them.  We only
        # track them if the file has rows.
        for table in files_database.tables.values():
            if not table.tracking_rows:
                continue

            if not table.name in live_database.tables:
                continue

            live_database.read_rows(table)

        mygrate = mygration(files_database, live_database, False)

        ops = []
        if mygrate.operations:
            ops.extend(mygrate.operations)

        rows = row_mygration(files_database, live_database)
        if rows.operations:
            ops.extend(rows.operations)

        if not ops:
            return []

        return [
            disable_checks(),
            *ops,
            enable_checks(),
        ]
